# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
import re
from copy import deepcopy
from functools import partial
from typing import List

import numpy as np
from PIL import Image, ImageDraw, ImageFont

from ....utils.fonts import PINGFANG_FONT
from ...common.result import (
    BaseCVResult,
    BaseResult,
    HtmlMixin,
    JsonMixin,
    LatexMixin,
    MarkdownMixin,
    WordMixin,
    XlsxMixin,
)
from .layout_objects import LayoutBlock
from .utils import get_seg_flag


def compile_title_pattern():
    # Precompiled regex pattern for matching numbering at the beginning of the title
    numbering_pattern = (
        r"(?:" + r"[1-9][0-9]*(?:\.[1-9][0-9]*)*[\.、]?|" + r"[\(\（](?:[1-9][0-9]*|["
        r"一二三四五六七八九十百千万亿零壹贰叁肆伍陆柒捌玖拾]+)[\)\）]|" + r"["
        r"一二三四五六七八九十百千万亿零壹贰叁肆伍陆柒捌玖拾]+"
        r"[、\.]?|" + r"(?:I|II|III|IV|V|VI|VII|VIII|IX|X)(?:\.|\s)" + r")"
    )
    return re.compile(r"^\s*(" + numbering_pattern + r")(\s*)(.*)$")


TITLE_RE_PATTERN = compile_title_pattern()


def format_title_func(block):
    """
    Normalize chapter title.
    Add the '#' to indicate the level of the title.
    If numbering exists, ensure there's exactly one space between it and the title content.
    If numbering does not exist, return the original title unchanged.

    :param title: Original chapter title string.
    :return: Normalized chapter title string.
    """
    title = block.content
    match = TITLE_RE_PATTERN.match(title)
    if match:
        numbering = match.group(1).strip()
        title_content = match.group(3).lstrip()
        # Return numbering and title content separated by one space
        title = numbering + " " + title_content

    title = title.rstrip(".")
    level = (
        title.count(
            ".",
        )
        + 1
        if "." in title
        else 1
    )
    return f"#{'#' * level} {title}".replace("-\n", "").replace(
        "\n",
        " ",
    )


def format_para_title_func(block):
    """
    Normalize chapter title.
    Add the '#' to indicate the level of the title.
    If numbering exists, ensure there's exactly one space between it and the title content.
    If numbering does not exist, return the original title unchanged.

    :param title: Original chapter title string.
    :return: Normalized chapter title string.
    """
    if not hasattr(block, "title_level"):
        return format_title_func(block)
    level = block.title_level
    title = block.content
    return f"#{'#' * level} {title}".replace("-\n", "").replace(
        "\n",
        " ",
    )


def format_centered_by_html(string, remove_symbol=True):
    if remove_symbol:
        string = string.replace("-\n", "").replace("\n", " ")
    return f'<div style="text-align: center;">{string}</div>' + "\n"


def format_text_plain_func(block):
    return block.content


def format_image_scaled_by_html_func(
    block, original_image_width, show_ocr_content=False
):
    img_tags = []
    if block.image is None:
        return ""
    image_path = block.image["path"]
    image_width = block.bbox[2] - block.bbox[0]
    scale = int(image_width / original_image_width * 100)
    img_tags.append(
        '<img src="{}" alt="Image" width="{}%" />'.format(
            image_path.replace("-\n", "").replace("\n", " "), scale
        ),
    )
    image_info = "\n".join(img_tags)
    if show_ocr_content:
        ocr_content = block.content
        image_info += "\n\n" + ocr_content + "\n\n"
    return image_info


def format_image_plain_func(block, show_ocr_content=False):
    img_tags = []
    if block.image:
        image_path = block.image["path"]
        img_tags.append(
            "![]({})".format(image_path.replace("-\n", "").replace("\n", " "))
        )
        image_info = "\n".join(img_tags)
        if show_ocr_content:
            ocr_content = block.content
            image_info += "\n\n" + ocr_content + "\n\n"
        return image_info
    return ""


def format_chart2table_func(block):
    lines_list = block.content.split("\n")
    column_num = len(lines_list[0].split("|"))
    lines_list.insert(1, "|".join(["---"] * column_num))
    lines_list = [f"|{line}|" for line in lines_list]
    return "\n".join(lines_list)


def simplify_table_func(table_code):
    return "\n" + table_code.replace("<html>", "").replace("</html>", "").replace(
        "<body>", ""
    ).replace("</body>", "")


def format_first_line_func(block, templates, format_func, spliter):
    lines = block.content.split(spliter)
    for idx in range(len(lines)):
        line = lines[idx]
        if line.strip() == "":
            continue
        if line.lower() in templates:
            lines[idx] = format_func(line)
        break
    return spliter.join(lines)


class LayoutParsingResultV2(
    BaseCVResult, HtmlMixin, XlsxMixin, MarkdownMixin, WordMixin, LatexMixin
):
    """Layout Parsing Result V2"""

    def __init__(self, data) -> None:
        """Initializes a new instance of the class with the specified data."""
        super().__init__(data)
        HtmlMixin.__init__(self)
        XlsxMixin.__init__(self)
        MarkdownMixin.__init__(self)
        JsonMixin.__init__(self)
        WordMixin.__init__(self)
        LatexMixin.__init__(self)

    def _to_img(self) -> dict[str, np.ndarray]:
        from .utils import get_show_color

        res_img_dict = {}
        model_settings = self["model_settings"]
        if model_settings["use_doc_preprocessor"]:
            for key, value in self["doc_preprocessor_res"].img.items():
                res_img_dict[key] = value
        res_img_dict["layout_det_res"] = self["layout_det_res"].img["res"]

        if model_settings["use_region_detection"]:
            res_img_dict["region_det_res"] = self["region_det_res"].img["res"]

        res_img_dict["overall_ocr_res"] = self["overall_ocr_res"].img["ocr_res_img"]

        if model_settings["use_table_recognition"] and len(self["table_res_list"]) > 0:
            table_cell_img = Image.fromarray(
                copy.deepcopy(self["doc_preprocessor_res"]["output_img"][:, :, ::-1])
            )
            table_draw = ImageDraw.Draw(table_cell_img)
            rectangle_color = (255, 0, 0)
            for sno in range(len(self["table_res_list"])):
                table_res = self["table_res_list"][sno]
                cell_box_list = table_res["cell_box_list"]
                for box in cell_box_list:
                    x1, y1, x2, y2 = [int(pos) for pos in box]
                    table_draw.rectangle(
                        [x1, y1, x2, y2], outline=rectangle_color, width=2
                    )
            res_img_dict["table_cell_img"] = table_cell_img

        if model_settings["use_seal_recognition"] and len(self["seal_res_list"]) > 0:
            for sno in range(len(self["seal_res_list"])):
                seal_res = self["seal_res_list"][sno]
                seal_region_id = seal_res["seal_region_id"]
                sub_seal_res_dict = seal_res.img
                key = f"seal_res_region{seal_region_id}"
                res_img_dict[key] = sub_seal_res_dict["ocr_res_img"]

        # for layout ordering image
        image = Image.fromarray(self["doc_preprocessor_res"]["output_img"][:, :, ::-1])
        draw = ImageDraw.Draw(image, "RGBA")
        font_size = int(0.018 * int(image.width)) + 2
        font = ImageFont.truetype(PINGFANG_FONT.path, font_size, encoding="utf-8")
        parsing_result: List[LayoutBlock] = self["parsing_res_list"]
        for block in parsing_result:
            bbox = block.bbox
            index = block.order_index
            label = block.label
            fill_color = get_show_color(label, False)
            draw.rectangle(bbox, fill=fill_color)
            if index is not None:
                text_position = (bbox[2] + 2, bbox[1] - font_size // 2)
                if int(image.width) - bbox[2] < font_size:
                    text_position = (
                        int(bbox[2] - font_size * 1.1),
                        bbox[1] - font_size // 2,
                    )
                draw.text(text_position, str(index), font=font, fill="red")

        res_img_dict["layout_order_res"] = image

        return res_img_dict

    def _to_str(self, *args, **kwargs) -> dict[str, str]:
        """Converts the instance's attributes to a dictionary and then to a string.

        Args:
            *args: Additional positional arguments passed to the base class method.
            **kwargs: Additional keyword arguments passed to the base class method.

        Returns:
            Dict[str, str]: A dictionary with the instance's attributes converted to strings.
        """
        data = {}
        data["input_path"] = self["input_path"]
        data["page_index"] = self["page_index"]
        data["page_count"] = self["page_count"]
        data["width"] = self["width"]
        data["height"] = self["height"]
        model_settings = self["model_settings"]
        data["model_settings"] = model_settings
        parsing_res_list: List[LayoutBlock] = self["parsing_res_list"]
        parsing_res_list = [
            {
                "block_label": parsing_res.label,
                "block_content": parsing_res.content,
                "block_bbox": parsing_res.bbox,
                "block_id": parsing_res.index,
                "block_order": parsing_res.order_index,
            }
            for parsing_res in parsing_res_list
        ]
        data["parsing_res_list"] = parsing_res_list
        if self["model_settings"]["use_doc_preprocessor"]:
            data["doc_preprocessor_res"] = self["doc_preprocessor_res"].str["res"]
        data["layout_det_res"] = self["layout_det_res"].str["res"]
        data["overall_ocr_res"] = self["overall_ocr_res"].str["res"]
        if model_settings["use_table_recognition"] and len(self["table_res_list"]) > 0:
            data["table_res_list"] = []
            for sno in range(len(self["table_res_list"])):
                table_res = self["table_res_list"][sno]
                data["table_res_list"].append(table_res.str["res"])
        if model_settings["use_seal_recognition"] and len(self["seal_res_list"]) > 0:
            data["seal_res_list"] = []
            for sno in range(len(self["seal_res_list"])):
                seal_res = self["seal_res_list"][sno]
                data["seal_res_list"].append(seal_res.str["res"])
        if (
            model_settings["use_formula_recognition"]
            and len(self["formula_res_list"]) > 0
        ):
            data["formula_res_list"] = []
            for sno in range(len(self["formula_res_list"])):
                formula_res = self["formula_res_list"][sno]
                data["formula_res_list"].append(formula_res.str["res"])

        return JsonMixin._to_str(data, *args, **kwargs)

    def _to_json(self, *args, **kwargs) -> dict[str, str]:
        """
        Converts the object's data to a JSON dictionary.

        Args:
            *args: Positional arguments passed to the JsonMixin._to_json method.
            **kwargs: Keyword arguments passed to the JsonMixin._to_json method.

        Returns:
            Dict[str, str]: A dictionary containing the object's data in JSON format.
        """
        if self["model_settings"].get("format_block_content", False):
            original_image_width = self["doc_preprocessor_res"]["output_img"].shape[1]
            format_text_func = lambda block: format_centered_by_html(
                format_text_plain_func(block)
            )
            format_image_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                )
            )

            if self["model_settings"].get("use_chart_recognition", False):
                format_chart_func = format_chart2table_func
            else:
                format_chart_func = format_image_func

            if self["model_settings"].get("use_seal_recognition", False):
                format_seal_func = lambda block: "\n".join(
                    [format_image_func(block), format_text_func(block)]
                )
            else:
                format_seal_func = format_image_func

            if self["model_settings"].get("use_table_recognition", False):
                format_table_func = lambda block: "\n" + format_text_func(
                    block
                ).replace("<table>", '<table border="1">')
            else:
                format_table_func = format_image_func

            if self["model_settings"].get("use_formula_recognition", False):
                format_formula_func = lambda block: f"$${block.content}$$"
            else:
                format_formula_func = format_image_func

            handle_funcs_dict = {
                "paragraph_title": format_para_title_func,
                "abstract_title": format_title_func,
                "reference_title": format_title_func,
                "content_title": format_title_func,
                "doc_title": lambda block: f"# {block.content}".replace(
                    "-\n",
                    "",
                ).replace("\n", " "),
                "table_title": format_text_func,
                "figure_title": format_text_func,
                "chart_title": format_text_func,
                "vision_footnote": lambda block: block.content.replace(
                    "\n\n", "\n"
                ).replace("\n", "\n\n"),
                "text": lambda block: block.content.replace("\n\n", "\n").replace(
                    "\n", "\n\n"
                ),
                "abstract": partial(
                    format_first_line_func,
                    templates=["摘要", "abstract"],
                    format_func=lambda l: f"## {l}\n",
                    spliter=" ",
                ),
                "content": lambda block: block.content.replace("-\n", "  \n").replace(
                    "\n", "  \n"
                ),
                "image": format_image_func,
                "chart": format_chart_func,
                "formula": format_formula_func,
                "table": format_table_func,
                "reference": partial(
                    format_first_line_func,
                    templates=["参考文献", "references"],
                    format_func=lambda l: f"## {l}",
                    spliter="\n",
                ),
                "algorithm": lambda block: block.content.strip("\n"),
                "seal": format_seal_func,
                "number": format_text_plain_func,
                "footnote": format_text_plain_func,
                "header": format_text_plain_func,
                "header_image": format_image_plain_func,
                "footer": format_text_plain_func,
                "footer_image": format_image_plain_func,
                "aside_text": format_text_plain_func,
            }
            for label in self["model_settings"].get("markdown_ignore_labels", []):
                handle_funcs_dict.pop(label, None)

        data = {}
        data["input_path"] = self["input_path"]
        data["page_index"] = self["page_index"]
        data["page_count"] = self["page_count"]
        data["width"] = self["width"]
        data["height"] = self["height"]
        model_settings = self["model_settings"]
        data["model_settings"] = model_settings
        parsing_res_list: List[LayoutBlock] = self["parsing_res_list"]
        parsing_res_list_json = []
        for parsing_res in parsing_res_list:
            res_dict = {
                "block_label": parsing_res.label,
                "block_content": parsing_res.content,
                "block_bbox": parsing_res.bbox,
                "block_id": parsing_res.index,
                "block_order": parsing_res.order_index,
            }
            if self["model_settings"].get("format_block_content", False):
                if handle_funcs_dict.get(parsing_res.label):
                    res_dict["block_content"] = handle_funcs_dict[parsing_res.label](
                        parsing_res
                    )
                else:
                    res_dict["block_content"] = parsing_res.content

            parsing_res_list_json.append(res_dict)
        data["parsing_res_list"] = parsing_res_list_json
        if self["model_settings"]["use_doc_preprocessor"]:
            data["doc_preprocessor_res"] = self["doc_preprocessor_res"].json["res"]
        data["layout_det_res"] = self["layout_det_res"].json["res"]
        data["overall_ocr_res"] = self["overall_ocr_res"].json["res"]
        if model_settings["use_table_recognition"] and len(self["table_res_list"]) > 0:
            data["table_res_list"] = []
            for sno in range(len(self["table_res_list"])):
                table_res = self["table_res_list"][sno]
                data["table_res_list"].append(table_res.json["res"])
        if model_settings["use_seal_recognition"] and len(self["seal_res_list"]) > 0:
            data["seal_res_list"] = []
            for sno in range(len(self["seal_res_list"])):
                seal_res = self["seal_res_list"][sno]
                data["seal_res_list"].append(seal_res.json["res"])
        if (
            model_settings["use_formula_recognition"]
            and len(self["formula_res_list"]) > 0
        ):
            data["formula_res_list"] = []
            for sno in range(len(self["formula_res_list"])):
                formula_res = self["formula_res_list"][sno]
                data["formula_res_list"].append(formula_res.json["res"])
        return JsonMixin._to_json(data, *args, **kwargs)

    def _to_html(self) -> dict[str, str]:
        """Converts the prediction to its corresponding HTML representation.

        Returns:
            Dict[str, str]: The str type HTML representation result.
        """
        model_settings = self["model_settings"]
        res_html_dict = {}
        if model_settings["use_table_recognition"] and len(self["table_res_list"]) > 0:
            for sno in range(len(self["table_res_list"])):
                table_res = self["table_res_list"][sno]
                table_region_id = table_res["table_region_id"]
                key = f"table_{table_region_id}"
                res_html_dict[key] = table_res.html["pred"]
        return res_html_dict

    def _to_xlsx(self) -> dict[str, str]:
        """Converts the prediction HTML to an XLSX file path.

        Returns:
            Dict[str, str]: The str type XLSX representation result.
        """
        model_settings = self["model_settings"]
        res_xlsx_dict = {}
        if model_settings["use_table_recognition"] and len(self["table_res_list"]) > 0:
            for sno in range(len(self["table_res_list"])):
                table_res = self["table_res_list"][sno]
                table_region_id = table_res["table_region_id"]
                key = f"table_{table_region_id}"
                res_xlsx_dict[key] = table_res.xlsx["pred"]
        return res_xlsx_dict

    def _to_markdown(self, pretty=True, show_formula_number=False) -> dict:
        """
        Save the parsing result to a Markdown file.

        Args:
            pretty (Optional[bool]): whether to pretty markdown by HTML, default by True.

        Returns:
            Dict
        """
        original_image_width = self["doc_preprocessor_res"]["output_img"].shape[1]

        if pretty:
            format_text_func = lambda block: format_centered_by_html(
                format_text_plain_func(block)
            )
            format_image_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                )
            )
        else:
            format_text_func = lambda block: block.content
            format_image_func = format_image_plain_func

        if self["model_settings"].get("use_chart_recognition", False):
            format_chart_func = format_chart2table_func
        else:
            format_chart_func = format_image_func

        if self["model_settings"].get("use_seal_recognition", False):
            format_seal_func = lambda block: "\n".join(
                [format_image_func(block), format_text_func(block)]
            )
        else:
            format_seal_func = format_image_func

        if self["model_settings"].get("use_table_recognition", False):
            if pretty:
                format_table_func = lambda block: "\n" + format_text_func(
                    block
                ).replace("<table>", '<table border="1">')
            else:
                format_table_func = lambda block: simplify_table_func(
                    "\n" + block.content
                )
        else:
            format_table_func = format_image_func

        if self["model_settings"].get("use_formula_recognition", False):
            format_formula_func = lambda block: f"$${block.content}$$"
        else:
            format_formula_func = format_image_func

        handle_funcs_dict = {
            "paragraph_title": format_title_func,
            "abstract_title": format_title_func,
            "reference_title": format_title_func,
            "content_title": format_title_func,
            "doc_title": lambda block: f"# {block.content}".replace(
                "-\n",
                "",
            ).replace("\n", " "),
            "table_title": format_text_func,
            "figure_title": format_text_func,
            "chart_title": format_text_func,
            "vision_footnote": lambda block: block.content.replace(
                "\n\n", "\n"
            ).replace("\n", "\n\n"),
            "text": lambda block: block.content.replace("\n\n", "\n").replace(
                "\n", "\n\n"
            ),
            "abstract": partial(
                format_first_line_func,
                templates=["摘要", "abstract"],
                format_func=lambda l: f"## {l}\n",
                spliter=" ",
            ),
            "content": lambda block: block.content.replace("-\n", "  \n").replace(
                "\n", "  \n"
            ),
            "image": format_image_func,
            "chart": format_chart_func,
            "formula": format_formula_func,
            "table": format_table_func,
            "reference": partial(
                format_first_line_func,
                templates=["参考文献", "references"],
                format_func=lambda l: f"## {l}",
                spliter="\n",
            ),
            "algorithm": lambda block: block.content.strip("\n"),
            "seal": format_seal_func,
            "number": format_text_plain_func,
            "footnote": format_text_plain_func,
            "header": format_text_plain_func,
            "header_image": format_image_plain_func,
            "footer": format_text_plain_func,
            "footer_image": format_image_plain_func,
            "aside_text": format_text_plain_func,
        }
        for label in self["model_settings"].get("markdown_ignore_labels", []):
            handle_funcs_dict.pop(label, None)

        markdown_content = ""
        last_label = None
        seg_start_flag = True
        seg_end_flag = True
        prev_block = None
        page_first_element_seg_start_flag = None
        page_last_element_seg_end_flag = None
        markdown_info = {}
        markdown_info["markdown_images"] = {}
        for block in self["parsing_res_list"]:
            seg_start_flag, seg_end_flag = get_seg_flag(block, prev_block)

            label = block.label
            if block.image is not None:
                markdown_info["markdown_images"][block.image["path"]] = block.image[
                    "img"
                ]
            page_first_element_seg_start_flag = (
                seg_start_flag
                if (page_first_element_seg_start_flag is None)
                else page_first_element_seg_start_flag
            )

            handle_func = handle_funcs_dict.get(label, None)
            if handle_func:
                prev_block = block
                if label == last_label == "text" and seg_start_flag == False:
                    markdown_content += handle_func(block)
                else:
                    markdown_content += (
                        "\n\n" + handle_func(block)
                        if markdown_content
                        else handle_func(block)
                    )
                last_label = label
        page_first_element_seg_start_flag = (
            True
            if page_first_element_seg_start_flag is None
            else page_first_element_seg_start_flag
        )
        page_last_element_seg_end_flag = seg_end_flag

        markdown_info["page_index"] = self["page_index"]
        markdown_info["input_path"] = self["input_path"]
        markdown_info["markdown_texts"] = markdown_content
        markdown_info["page_continuation_flags"] = (
            page_first_element_seg_start_flag,
            page_last_element_seg_end_flag,
        )
        for img in self["imgs_in_doc"]:
            markdown_info["markdown_images"][img["path"]] = img["img"]

        return markdown_info

    def _to_word(self) -> dict:
        from docx.enum.text import WD_ALIGN_PARAGRAPH

        """
        Convert the object's parsing result into a Word-compatible dict.

        Returns:
            dict: {
                "word_blocks": List[Dict],       # Simplified list of content blocks
                "original_image_width": int,   # Pixel width of the source page
                "input_path": str,             # Original input file path
                "images": List[Dict]           # List of {"path": str, "img": PIL.Image}
            }
        """

        word_blocks = []
        image = []

        STYLE_MAP = {
            "doc_title": {
                "level": 0,
                "size": 20,
                "bold": True,
                "align": WD_ALIGN_PARAGRAPH.CENTER,
            },
            "header": {
                "size": 16,
                "bold": True,
                "align": WD_ALIGN_PARAGRAPH.CENTER,
            },
            "abstract_title": {
                "level": 1,
                "size": 14,
                "bold": True,
                "align": WD_ALIGN_PARAGRAPH.CENTER,
            },
            "content_title": {
                "level": 1,
                "size": 14,
                "bold": True,
                "align": WD_ALIGN_PARAGRAPH.LEFT,
            },
            "reference_title": {
                "level": 1,
                "size": 14,
                "bold": True,
                "align": WD_ALIGN_PARAGRAPH.LEFT,
            },
            "paragraph_title": {
                "level": 2,
                "size": 14,
                "bold": True,
                "align": WD_ALIGN_PARAGRAPH.LEFT,
            },
            "abstract": {"size": 12, "align": WD_ALIGN_PARAGRAPH.JUSTIFY},
            "text": {
                "size": 12,
                "align": WD_ALIGN_PARAGRAPH.JUSTIFY,
                "indent": True,
            },
            "figure_title": {"size": 10, "align": WD_ALIGN_PARAGRAPH.CENTER},
            "table_title": {"size": 10, "align": WD_ALIGN_PARAGRAPH.CENTER},
            "chart_title": {"size": 10, "align": WD_ALIGN_PARAGRAPH.CENTER},
            "reference": {"size": 12, "align": WD_ALIGN_PARAGRAPH.JUSTIFY},
            "algorithm": {
                "font": "Courier New",
                "size": 11,
                "align": WD_ALIGN_PARAGRAPH.LEFT,
            },
            "formula": {"size": 12, "align": WD_ALIGN_PARAGRAPH.CENTER},
            "vision_footnote": {"size": 9, "align": WD_ALIGN_PARAGRAPH.LEFT},
            "number": {"size": 9, "align": WD_ALIGN_PARAGRAPH.CENTER},
            "footer": {"size": 9, "align": WD_ALIGN_PARAGRAPH.CENTER},
        }

        for block in self["parsing_res_list"]:

            label = block.label
            content = getattr(block, "content", "")
            if label in ["image", "chart", "seal"]:
                content = block.image["path"]
            config = STYLE_MAP.get(
                label,
                {"size": 12, "align": WD_ALIGN_PARAGRAPH.LEFT, "indent": True},
            )
            block_dict = {
                "type": label,
                "content": deepcopy(content),
                "config": config,
            }
            word_blocks.append(block_dict)
            if block.image is not None:
                image.append({"path": block.image["path"], "img": block.image["img"]})

        return {
            "word_blocks": word_blocks,
            "original_image_width": self["doc_preprocessor_res"]["output_img"].shape[1],
            "input_path": self["input_path"],
            "images": image,
        }

    def _to_latex(self) -> dict:
        """
        Convert the object's parsing result into a latex-compatible dict.

        Returns:
            dict: {
                "latex_blocks": List[Dict],       # Simplified list of content blocks
                "input_path": str,             # Original input file path
                "images": List[Dict]           # List of {"path": str, "img": PIL.Image}
            }
        """
        latex_blocks = []
        image = []

        for block in self["parsing_res_list"]:

            label = block.label
            content = getattr(block, "content", "")
            if label in ["image", "chart", "seal"]:
                content = block.image["path"]
            block_dict = {
                "type": label,
                "content": deepcopy(content),
            }
            latex_blocks.append(block_dict)
            if block.image is not None:
                image.append({"path": block.image["path"], "img": block.image["img"]})

        return {
            "latex_blocks": latex_blocks,
            "images": image,
            "input_path": self["input_path"],
        }


class ProcessedLayoutParsingResult(BaseResult, MarkdownMixin):

    def __init__(self, data) -> None:
        """Initializes a new instance of the class with the specified data."""
        super().__init__(data)
        MarkdownMixin.__init__(self)

    def _to_markdown(self, pretty=True, show_formula_number=False) -> dict:
        """
        Save the parsing result to a Markdown file.

        Args:
            pretty (Optional[bool]): whether to pretty markdown by HTML, default by True.

        Returns:
            Dict
        """

        self["model_settings"] = self["model_settings"][0]
        self["input_path"] = self["input_path"][0]
        self["doc_preprocessor_res"] = self["doc_preprocessor_res"][0]
        self["page_index"] = None
        self["width"] = None
        self["height"] = None
        self["page_count"] = None

        original_image_width = self["doc_preprocessor_res"]["output_img"].shape[1]

        if pretty:
            format_text_func = lambda block: format_centered_by_html(
                format_text_plain_func(block)
            )
            format_image_func = lambda block: format_centered_by_html(
                format_image_scaled_by_html_func(
                    block,
                    original_image_width=original_image_width,
                )
            )
        else:
            format_text_func = lambda block: block.content
            format_image_func = format_image_plain_func

        if self["model_settings"].get("use_chart_recognition", False):
            format_chart_func = format_chart2table_func
        else:
            format_chart_func = format_image_func

        if self["model_settings"].get("use_seal_recognition", False):
            format_seal_func = lambda block: "\n".join(
                [format_image_func(block), format_text_func(block)]
            )
        else:
            format_seal_func = format_image_func

        if self["model_settings"].get("use_table_recognition", False):
            if pretty:
                format_table_func = lambda block: "\n" + format_text_func(
                    block
                ).replace("<table>", '<table border="1">')
            else:
                format_table_func = lambda block: simplify_table_func(
                    "\n" + block.content
                )
        else:
            format_table_func = format_image_func

        if self["model_settings"].get("use_formula_recognition", False):
            format_formula_func = lambda block: f"$${block.content}$$"
        else:
            format_formula_func = format_image_func

        handle_funcs_dict = {
            "paragraph_title": format_para_title_func,
            "abstract_title": format_title_func,
            "reference_title": format_title_func,
            "content_title": format_title_func,
            "doc_title": lambda block: f"# {block.content}".replace(
                "-\n",
                "",
            ).replace("\n", " "),
            "table_title": format_text_func,
            "figure_title": format_text_func,
            "chart_title": format_text_func,
            "vision_footnote": lambda block: block.content.replace(
                "\n\n", "\n"
            ).replace("\n", "\n\n"),
            "text": lambda block: block.content.replace("\n\n", "\n").replace(
                "\n", "\n\n"
            ),
            "abstract": partial(
                format_first_line_func,
                templates=["摘要", "abstract"],
                format_func=lambda l: f"## {l}\n",
                spliter=" ",
            ),
            "content": lambda block: block.content.replace("-\n", "  \n").replace(
                "\n", "  \n"
            ),
            "image": format_image_func,
            "chart": format_chart_func,
            "formula": format_formula_func,
            "table": format_table_func,
            "reference": partial(
                format_first_line_func,
                templates=["参考文献", "references"],
                format_func=lambda l: f"## {l}",
                spliter="\n",
            ),
            "algorithm": lambda block: block.content.strip("\n"),
            "seal": format_seal_func,
        }
        for label in self["model_settings"].get("markdown_ignore_labels", []):
            handle_funcs_dict.pop(label, None)

        markdown_content = ""
        markdown_info = {}
        markdown_info["markdown_images"] = {}
        pages_list = self["parsing_res_list"]
        global_block_id = 0

        for page_blocks in pages_list:

            if not page_blocks:
                continue
            for idx, block in enumerate(page_blocks):

                label = block.label

                if block.image is not None:
                    markdown_info["markdown_images"][block.image["path"]] = block.image[
                        "img"
                    ]

                handle_func = handle_funcs_dict.get(label, None)

                if handle_func:
                    current_text = handle_func(block)

                    if markdown_content:
                        markdown_content += "\n\n" + current_text
                    else:
                        markdown_content += current_text

                if block.group_id is None:
                    block.group_id = global_block_id
                global_block_id += 1

        markdown_info["page_index"] = self["page_index"]
        markdown_info["input_path"] = self["input_path"]
        markdown_info["markdown_texts"] = markdown_content
        for img in self["imgs_in_doc"]:
            markdown_info["markdown_images"][img["path"]] = img["img"]

        return markdown_info



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import importlib
from pathlib import Path
from typing import Any, Dict, Optional, Type

from packaging.version import Version
from pydantic import BaseModel

from ....utils import logging
from ...utils.official_models import official_models
from ..utils import check_backend, model_name_to_module_name

NETWORK_CLASS_GETTER_KEY = "get_network_class"
PROCESSOR_CLASS_GETTER_KEY = "get_processor_class"
CONFIG_GETTER_KEY = "get_config"
CHAT_TEMPLATE_PATH_GETTER_KEY = "get_chat_template_path"
DEFAULT_CHAT_TEMPLATE_FILENAME = "chat_template.jinja"

ALL_MODEL_INFO = {
    "PaddleOCR-VL-0.9B": {
        "min_vllm_version": "0.11.1",
        "min_sglang_version": "0.5.7",
    },
    "PaddleOCR-VL-1.5-0.9B": {
        "min_vllm_version": "0.11.1",
        "min_sglang_version": "0.5.7",
    },
}


def _check_model_name_and_backend(model_name, backend):
    if model_name not in ALL_MODEL_INFO:
        raise ValueError(f"Unknown model: {model_name}")

    check_backend(backend)


def is_integrated_model_available(model_name, backend):
    _check_model_name_and_backend(model_name, backend)

    model_info = ALL_MODEL_INFO[model_name]

    if f"min_{backend}_version" in model_info:
        if backend == "vllm":
            import vllm

            backend_lib_version = vllm.__version__
        elif backend == "sglang":
            import sglang

            backend_lib_version = sglang.__version__
        else:
            backend_lib_version = "0.0.0"
        return Version(backend_lib_version) >= Version(
            model_info[f"min_{backend}_version"]
        )

    return False


def get_model_dir(model_name, backend):
    _check_model_name_and_backend(model_name, backend)

    try:
        model_dir = official_models[model_name]
    except Exception as e:
        raise RuntimeError(
            f"Could not prepare the official model for the {repr(model_name)} model with the {repr(backend)} backend."
        ) from e

    return str(model_dir)


def get_model_components(model_name, backend):
    def _get_component(getter_key):
        if not hasattr(model_module, getter_key):
            raise RuntimeError(f"`{model_module}` does not have `{getter_key}`")
        getter = getattr(model_module, getter_key)
        comp = getter(backend)
        return comp

    _check_model_name_and_backend(model_name, backend)

    mod_name = model_name_to_module_name(model_name)

    try:
        model_module = importlib.import_module(f".{mod_name}", package=__package__)
    except ModuleNotFoundError as e:
        raise ValueError(f"Unknown model: {model_name}") from e

    network_class = _get_component(NETWORK_CLASS_GETTER_KEY)

    if backend == "sglang":
        processor_class = _get_component(PROCESSOR_CLASS_GETTER_KEY)
    else:
        processor_class = None

    return network_class, processor_class


def get_default_config(model_name, backend):
    _check_model_name_and_backend(model_name, backend)

    mod_name = model_name_to_module_name(model_name)

    try:
        config_module = importlib.import_module(
            f"..configs.{mod_name}", package=__package__
        )
    except ModuleNotFoundError:
        logging.debug("No default configs were found for the model '%s'", model_name)
        default_config = {}
    else:
        if not hasattr(config_module, CONFIG_GETTER_KEY):
            raise RuntimeError(f"`{config_module}` does not have `{CONFIG_GETTER_KEY}`")
        config_getter = getattr(config_module, CONFIG_GETTER_KEY)
        default_config = config_getter(backend)

    return default_config


@contextlib.contextmanager
def get_chat_template_path(model_name, backend, model_dir):
    _check_model_name_and_backend(model_name, backend)

    with importlib.resources.path(
        "paddlex.inference.genai.chat_templates", f"{model_name}.jinja"
    ) as chat_template_path:
        if not chat_template_path.exists():
            default_chat_template_path = Path(model_dir, DEFAULT_CHAT_TEMPLATE_FILENAME)
            if (
                default_chat_template_path.exists()
                and default_chat_template_path.is_file()
            ):
                # TODO: Support symbolic links
                yield default_chat_template_path
            else:
                logging.debug(
                    "No chat template was found for the model '%s' with the backend '%s'",
                    model_name,
                    backend,
                )
                yield None
        else:
            yield chat_template_path



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This file referred to github.com/onnx/onnx.git

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil

TOP_DIR = os.path.realpath(os.path.dirname(__file__))
TOP_DIR = os.path.split(TOP_DIR)[0]
PACKAGE_NAME = os.getenv("PACKAGE_NAME", "ultra_infer")
wheel_name = os.getenv("WHEEL_NAME", "ultra-infer-python")

if not os.path.exists(PACKAGE_NAME):
    shutil.copytree("ultra_infer", PACKAGE_NAME)

import glob
import multiprocessing
import platform
import shlex
import subprocess
import sys
from collections import namedtuple
from contextlib import contextmanager
from distutils import log, sysconfig
from distutils.spawn import find_executable
from textwrap import dedent

import setuptools
import setuptools.command.build_ext
import setuptools.command.build_py
import setuptools.command.develop

with open(os.path.join(TOP_DIR, "python", "requirements.txt")) as fin:
    REQUIRED_PACKAGES = fin.read()

if os.getenv("BUILD_ON_CPU", "OFF") == "ON":
    os.environ["ENABLE_PADDLE_BACKEND"] = "ON"
    os.environ["ENABLE_ORT_BACKEND"] = "ON"
    os.environ["ENABLE_OPENVINO_BACKEND"] = "ON"
    os.environ["ENABLE_VISION"] = "ON"
    os.environ["ENABLE_TEXT"] = "ON"
    os.environ["WITH_GPU"] = "OFF"

setup_configs = dict()
setup_configs["LIBRARY_NAME"] = PACKAGE_NAME
setup_configs["PY_LIBRARY_NAME"] = PACKAGE_NAME + "_main"
# Backend options
setup_configs["ENABLE_TVM_BACKEND"] = os.getenv("ENABLE_TVM_BACKEND", "OFF")
setup_configs["ENABLE_RKNPU2_BACKEND"] = os.getenv("ENABLE_RKNPU2_BACKEND", "OFF")
setup_configs["ENABLE_SOPHGO_BACKEND"] = os.getenv("ENABLE_SOPHGO_BACKEND", "OFF")
setup_configs["ENABLE_ORT_BACKEND"] = os.getenv("ENABLE_ORT_BACKEND", "OFF")
setup_configs["ENABLE_OPENVINO_BACKEND"] = os.getenv("ENABLE_OPENVINO_BACKEND", "OFF")
setup_configs["ENABLE_PADDLE_BACKEND"] = os.getenv("ENABLE_PADDLE_BACKEND", "OFF")
setup_configs["ENABLE_POROS_BACKEND"] = os.getenv("ENABLE_POROS_BACKEND", "OFF")
setup_configs["ENABLE_TRT_BACKEND"] = os.getenv("ENABLE_TRT_BACKEND", "OFF")
setup_configs["ENABLE_LITE_BACKEND"] = os.getenv("ENABLE_LITE_BACKEND", "OFF")
setup_configs["ENABLE_OM_BACKEND"] = os.getenv("ENABLE_OM_BACKEND", "OFF")
setup_configs["ENABLE_PADDLE2ONNX"] = os.getenv("ENABLE_PADDLE2ONNX", "OFF")
setup_configs["ENABLE_VISION"] = os.getenv("ENABLE_VISION", "OFF")
setup_configs["ENABLE_FLYCV"] = os.getenv("ENABLE_FLYCV", "OFF")
setup_configs["ENABLE_CVCUDA"] = os.getenv("ENABLE_CVCUDA", "OFF")
setup_configs["ENABLE_TEXT"] = os.getenv("ENABLE_TEXT", "OFF")
setup_configs["ENABLE_BENCHMARK"] = os.getenv("ENABLE_BENCHMARK", "OFF")
# Hardware options
setup_configs["WITH_GPU"] = os.getenv("WITH_GPU", "OFF")
setup_configs["WITH_IPU"] = os.getenv("WITH_IPU", "OFF")
setup_configs["WITH_OPENCL"] = os.getenv("WITH_OPENCL", "OFF")
setup_configs["WITH_TIMVX"] = os.getenv("WITH_TIMVX", "OFF")
setup_configs["WITH_DIRECTML"] = os.getenv("WITH_DIRECTML", "OFF")
setup_configs["WITH_ASCEND"] = os.getenv("WITH_ASCEND", "OFF")
setup_configs["WITH_KUNLUNXIN"] = os.getenv("WITH_KUNLUNXIN", "OFF")
setup_configs["RKNN2_TARGET_SOC"] = os.getenv("RKNN2_TARGET_SOC", "")
setup_configs["DEVICE_TYPE"] = os.getenv("DEVICE_TYPE", "")
# Custom deps settings
setup_configs["TRT_DIRECTORY"] = os.getenv("TRT_DIRECTORY", "UNDEFINED")
setup_configs["CUDA_DIRECTORY"] = os.getenv("CUDA_DIRECTORY", "/usr/local/cuda")
setup_configs["OPENCV_DIRECTORY"] = os.getenv("OPENCV_DIRECTORY", "")
setup_configs["ORT_DIRECTORY"] = os.getenv("ORT_DIRECTORY", "")
setup_configs["OPENVINO_DIRECTORY"] = os.getenv("OPENVINO_DIRECTORY", "")
setup_configs["PADDLEINFERENCE_DIRECTORY"] = os.getenv("PADDLEINFERENCE_DIRECTORY", "")
setup_configs["PADDLEINFERENCE_VERSION"] = os.getenv("PADDLEINFERENCE_VERSION", "")
setup_configs["PADDLEINFERENCE_URL"] = os.getenv("PADDLEINFERENCE_URL", "")
setup_configs["PADDLEINFERENCE_API_COMPAT_2_4_x"] = os.getenv(
    "PADDLEINFERENCE_API_COMPAT_2_4_x", "OFF"
)
setup_configs["PADDLEINFERENCE_API_COMPAT_2_5_x"] = os.getenv(
    "PADDLEINFERENCE_API_COMPAT_2_5_x", "OFF"
)
setup_configs["PADDLEINFERENCE_API_COMPAT_2_6_x"] = os.getenv(
    "PADDLEINFERENCE_API_COMPAT_2_6_x", "OFF"
)
setup_configs["PADDLEINFERENCE_API_COMPAT_DEV"] = os.getenv(
    "PADDLEINFERENCE_API_COMPAT_DEV", "OFF"
)
setup_configs["PADDLEINFERENCE_API_CUSTOM_OP"] = os.getenv(
    "PADDLEINFERENCE_API_CUSTOM_OP", "OFF"
)
setup_configs["PADDLE2ONNX_URL"] = os.getenv("PADDLE2ONNX_URL", "")
setup_configs["PADDLELITE_URL"] = os.getenv("PADDLELITE_URL", "")

# Other settings
setup_configs["BUILD_ON_JETSON"] = os.getenv("BUILD_ON_JETSON", "OFF")
setup_configs["BUILD_PADDLE2ONNX"] = os.getenv("BUILD_PADDLE2ONNX", "OFF")

if setup_configs["RKNN2_TARGET_SOC"] != "" or setup_configs["BUILD_ON_JETSON"] != "OFF":
    REQUIRED_PACKAGES = REQUIRED_PACKAGES.replace("opencv-contrib-python", "")

if wheel_name == "ultra-infer-python":
    device_type = setup_configs["DEVICE_TYPE"]
    if device_type:
        if device_type not in ["GPU", "IPU", "NPU"]:
            sys.exit(
                f"Invalid DEVICE_TYPE: '{device_type}'. Supported values are: GPU, IPU, NPU. "
                "Please update the DEVICE_TYPE environment variable accordingly."
            )
        wheel_name = f"ultra-infer-{device_type.lower()}-python"
    else:
        if (
            setup_configs["WITH_GPU"] == "ON"
            or setup_configs["BUILD_ON_JETSON"] == "ON"
        ):
            wheel_name = "ultra-infer-gpu-python"
        elif setup_configs["WITH_IPU"] == "ON":
            wheel_name = "ultra-infer-ipu-python"

if os.getenv("CMAKE_CXX_COMPILER", None) is not None:
    setup_configs["CMAKE_CXX_COMPILER"] = os.getenv("CMAKE_CXX_COMPILER")

SRC_DIR = os.path.join(TOP_DIR, PACKAGE_NAME)
PYTHON_SRC_DIR = os.path.join(TOP_DIR, "python", PACKAGE_NAME)
CMAKE_BUILD_DIR = os.path.join(TOP_DIR, "python", ".setuptools-cmake-build")

WINDOWS = os.name == "nt"

CMAKE = find_executable("cmake3") or find_executable("cmake")
MAKE = find_executable("make")

setup_requires = []
extras_require = {}

################################################################################
# Global variables for controlling the build variant
################################################################################

# Default value is set to TRUE\1 to keep the settings same as the current ones.
# However going forward the recomemded way to is to set this to False\0
USE_MSVC_STATIC_RUNTIME = bool(os.getenv("USE_MSVC_STATIC_RUNTIME", "1") == "1")
ONNX_NAMESPACE = os.getenv("ONNX_NAMESPACE", "paddle2onnx")
################################################################################
# Version
################################################################################

try:
    git_version = (
        subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=TOP_DIR)
        .decode("ascii")
        .strip()
    )
except (OSError, subprocess.CalledProcessError):
    git_version = None

extra_version_info = ""
if setup_configs["PADDLEINFERENCE_VERSION"] != "":
    extra_version_info += "." + setup_configs["PADDLEINFERENCE_VERSION"]

with open(os.path.join(TOP_DIR, "VERSION_NUMBER")) as version_file:
    VersionInfo = namedtuple(
        "VersionInfo",
        [
            "version",
            "git_version",
            "extra_version_info",
            "enable_trt_backend",
            "enable_paddle_backend",
            "with_gpu",
        ],
    )(
        version=version_file.read().strip(),
        git_version=git_version,
        extra_version_info=extra_version_info.strip("."),
        enable_trt_backend=setup_configs["ENABLE_TRT_BACKEND"],
        enable_paddle_backend=setup_configs["ENABLE_PADDLE_BACKEND"],
        with_gpu=setup_configs["WITH_GPU"],
    )

################################################################################
# Pre Check
################################################################################

assert CMAKE, 'Could not find "cmake" executable!'

################################################################################
# Utilities
################################################################################


@contextmanager
def cd(path):
    if not os.path.isabs(path):
        raise RuntimeError("Can only cd to absolute path, got: {}".format(path))
    orig_path = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(orig_path)


################################################################################
# Customized commands
################################################################################


class NoOptionCommand(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


def get_all_files(dirname):
    files = list()
    for root, dirs, filenames in os.walk(dirname):
        for f in filenames:
            fullname = os.path.join(root, f)
            files.append(fullname)
    return files


class create_version(NoOptionCommand):
    def run(self):
        with open(os.path.join(PYTHON_SRC_DIR, "code_version.py"), "w") as f:
            f.write(
                dedent(
                    """\
            # This file is generated by setup.py. DO NOT EDIT!
            from __future__ import absolute_import
            from __future__ import division
            from __future__ import print_function
            from __future__ import unicode_literals
            version = '{version}'
            git_version = '{git_version}'
            extra_version_info = '{extra_version_info}'
            enable_trt_backend = '{enable_trt_backend}'
            enable_paddle_backend = '{enable_paddle_backend}'
            with_gpu = '{with_gpu}'
            """.format(
                        **dict(VersionInfo._asdict())
                    )
                )
            )


class cmake_build(setuptools.Command):
    """
    Compiles everything when `python setupmnm.py build` is run using cmake.
    Custom args can be passed to cmake by specifying the `CMAKE_ARGS`
    environment variable.
    The number of CPUs used by `make` can be specified by passing `-j<ncpus>`
    to `setup.py build`.  By default all CPUs are used.
    """

    user_options = [
        (str("jobs="), str("j"), str("Specifies the number of jobs to use with make"))
    ]

    built = False

    def initialize_options(self):
        self.jobs = None

    def finalize_options(self):
        if sys.version_info[0] >= 3:
            self.set_undefined_options("build", ("parallel", "jobs"))
        if self.jobs is None and os.getenv("MAX_JOBS") is not None:
            self.jobs = os.getenv("MAX_JOBS")
        self.jobs = multiprocessing.cpu_count() if self.jobs is None else int(self.jobs)

    def run(self):
        if cmake_build.built:
            return
        cmake_build.built = True
        if not os.path.exists(CMAKE_BUILD_DIR):
            os.makedirs(CMAKE_BUILD_DIR)

        with cd(CMAKE_BUILD_DIR):
            build_type = "Release"
            # configure
            cmake_args = [
                CMAKE,
                "-DPYTHON_INCLUDE_DIR={}".format(sysconfig.get_python_inc()),
                "-DPYTHON_EXECUTABLE={}".format(sys.executable),
                "-DBUILD_ULTRAINFER_PYTHON=ON",
                "-DCMAKE_EXPORT_COMPILE_COMMANDS=ON",
                "-DONNX_NAMESPACE={}".format(ONNX_NAMESPACE),
                "-DPY_EXT_SUFFIX={}".format(
                    sysconfig.get_config_var("EXT_SUFFIX") or ""
                ),
                "-DCUDA_VERSION={}".format(os.getenv("CUDA_VERSION")),
            ]
            cmake_args.append("-DCMAKE_BUILD_TYPE=%s" % build_type)
            for k, v in setup_configs.items():
                cmake_args.append("-D{}={}".format(k, v))
            if WINDOWS:
                cmake_args.extend(
                    [
                        # we need to link with libpython on windows, so
                        # passing python version to window in order to
                        # find python in cmake
                        "-DPY_VERSION={}".format(
                            "{0}.{1}".format(*sys.version_info[:2])
                        ),
                    ]
                )
                if platform.architecture()[0] == "64bit":
                    cmake_args.extend(["-A", "x64", "-T", "host=x64"])
                else:
                    cmake_args.extend(["-A", "Win32", "-T", "host=x86"])
            if "CMAKE_ARGS" in os.environ:
                extra_cmake_args = shlex.split(os.environ["CMAKE_ARGS"])
                # prevent crossfire with downstream scripts
                del os.environ["CMAKE_ARGS"]
                log.info("Extra cmake args: {}".format(extra_cmake_args))
                cmake_args.extend(extra_cmake_args)
            cmake_args.append(TOP_DIR)
            subprocess.check_call(cmake_args)

            build_args = [CMAKE, "--build", os.curdir]
            if WINDOWS:
                build_args.extend(["--config", build_type])
                build_args.extend(["--", "/maxcpucount:{}".format(self.jobs)])
            else:
                build_args.extend(["--", "-j", str(self.jobs)])
            subprocess.check_call(build_args)


class build_py(setuptools.command.build_py.build_py):
    def run(self):
        self.run_command("create_version")
        self.run_command("cmake_build")

        generated_python_files = glob.glob(
            os.path.join(CMAKE_BUILD_DIR, PACKAGE_NAME, "*.py")
        ) + glob.glob(os.path.join(CMAKE_BUILD_DIR, PACKAGE_NAME, "*.pyi"))

        for src in generated_python_files:
            dst = os.path.join(TOP_DIR, os.path.relpath(src, CMAKE_BUILD_DIR))
            self.copy_file(src, dst)

        return setuptools.command.build_py.build_py.run(self)


class develop(setuptools.command.develop.develop):
    def run(self):
        self.run_command("build_py")
        setuptools.command.develop.develop.run(self)


class build_ext(setuptools.command.build_ext.build_ext):
    def run(self):
        self.run_command("cmake_build")
        setuptools.command.build_ext.build_ext.run(self)

    def build_extensions(self):
        for ext in self.extensions:
            fullname = self.get_ext_fullname(ext.name)
            filename = os.path.basename(self.get_ext_filename(fullname))

            lib_path = CMAKE_BUILD_DIR
            if os.name == "nt":
                debug_lib_dir = os.path.join(lib_path, "Debug")
                release_lib_dir = os.path.join(lib_path, "Release")
                if os.path.exists(debug_lib_dir):
                    lib_path = debug_lib_dir
                elif os.path.exists(release_lib_dir):
                    lib_path = release_lib_dir
            src = os.path.join(lib_path, filename)
            dst = os.path.join(os.path.realpath(self.build_lib), PACKAGE_NAME, filename)
            self.copy_file(src, dst)


cmdclass = {
    "create_version": create_version,
    "cmake_build": cmake_build,
    "build_py": build_py,
    "develop": develop,
    "build_ext": build_ext,
}

################################################################################
# Extensions
################################################################################

ext_modules = [
    setuptools.Extension(
        name=str(PACKAGE_NAME + "." + setup_configs["PY_LIBRARY_NAME"]), sources=[]
    ),
]

################################################################################
# Packages
################################################################################

# no need to do fancy stuff so far
if PACKAGE_NAME != "ultra_infer":
    packages = setuptools.find_packages(exclude=["ultra_infer*", "scripts"])
else:
    packages = setuptools.find_packages(exclude=["xencrypt*", "scripts"])

################################################################################
# Test
################################################################################

if sys.version_info[0] == 3:
    # Mypy doesn't work with Python 2
    extras_require["mypy"] = ["mypy==0.600"]

################################################################################
# Pyonly
################################################################################

extras_require["pyonly"] = [
    "pillow<10.0.0",
    "pandas>=0.25.0,<=1.3.5",
    "pycocotools",
    "matplotlib",
    "chinese_calendar",
    "joblib",
    "scikit-image",
    "scikit-learn>=1.3.2",
    "tokenizers",
]

################################################################################
# Final
################################################################################

package_data = {PACKAGE_NAME: ["LICENSE", "ThirdPartyNotices.txt"]}

if sys.argv[1] == "install" or sys.argv[1] == "bdist_wheel":
    shutil.copy(
        os.path.join(TOP_DIR, "ThirdPartyNotices.txt"),
        os.path.join(TOP_DIR, "python", PACKAGE_NAME),
    )
    shutil.copy(
        os.path.join(TOP_DIR, "LICENSE"), os.path.join(TOP_DIR, "python", PACKAGE_NAME)
    )
    if not os.path.exists(
        os.path.join(TOP_DIR, "python", PACKAGE_NAME, "libs", "third_libs")
    ):
        print(
            f"Didn't detect path: {PACKAGE_NAME}/libs/third_libs exist, please execute `python setup.py build` first"
        )
        sys.exit(0)
    from scripts.process_libraries import process_libraries

    all_lib_data = process_libraries(os.path.split(os.path.abspath(__file__))[0])
    package_data[PACKAGE_NAME].extend(all_lib_data)
    setuptools.setup(
        name=wheel_name,
        version=VersionInfo.version + extra_version_info,
        ext_modules=ext_modules,
        description="Deploy Kit Tool For Deeplearning models.",
        packages=packages,
        package_data=package_data,
        include_package_data=True,
        setup_requires=setup_requires,
        extras_require=extras_require,
        author="ultra_infer",
        install_requires=REQUIRED_PACKAGES,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
        license="Apache 2.0",
    )
else:
    setuptools.setup(
        name=wheel_name,
        version=VersionInfo.version + extra_version_info,
        description="Deploy Kit Tool For Deeplearning models.",
        ext_modules=ext_modules,
        cmdclass=cmdclass,
        packages=packages,
        package_data=package_data,
        include_package_data=False,
        setup_requires=setup_requires,
        extras_require=extras_require,
        author="ultra_infer",
        install_requires=REQUIRED_PACKAGES,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
        license="Apache 2.0",
    )



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common fusion operators.
"""

# TODO: Support XPU

import paddle
import paddle.nn.functional as F
from paddle import matmul, tensor
from paddle.nn import Linear
from paddle.nn.functional.flash_attention import flashmask_attention

__all__ = [
    "matmul",
    "Linear",
]


def _fusion_flash_attention(
    q,
    k,
    v,
    training_mode,
    attention_probs_dropout_prob,
    use_sparse_flash_attn,
    attention_mask=None,
    attn_mask_start_row_indices=None,
    rr_flash_attn=None,
):
    """
    Performs fused flash attention with multiple implementation variants.

    Args:
        q (paddle.Tensor): Query tensor with shape [batch, heads, seq_len, dim_head]
        k (paddle.Tensor): Key tensor with shape [batch, heads, seq_len, dim_head]
        v (paddle.Tensor): Value tensor with shape [batch, heads, seq_len, dim_head]
        training_mode (bool): Whether in training mode (affects dropout)
        attention_probs_dropout_prob (float): Dropout probability for attention weights
        use_sparse_flash_attn (bool): Whether to use sparse flash attention optimization
        attention_mask (Optional[paddle.Tensor]): Dense attention mask (default: None)
        attn_mask_start_row_indices (Optional[paddle.Tensor]): Sparse mask indices (default: None)
        rr_flash_attn (Optional[Callable]): Recomputation wrapper for flash attention (default: None)

    Returns:
        Tuple[paddle.Tensor, Optional[paddle.Tensor]]:
            - Output tensor with shape [batch, seq_len, heads*dim_head]
            - Attention weights (None for flash attention implementations)

    Raises:
        Warning: If sparse flash attention is requested but unavailable
        ValueError: If invalid combination of mask inputs is provided
    """

    version = paddle.version.full_version
    if attn_mask_start_row_indices is not None:
        if use_sparse_flash_attn:
            if rr_flash_attn is None:
                out = flashmask_attention(
                    q,
                    k,
                    v,
                    startend_row_indices=attn_mask_start_row_indices.unsqueeze(-1),
                    causal=True,
                )
            else:
                out = rr_flash_attn(
                    flashmask_attention,
                    q,
                    k,
                    v,
                    startend_row_indices=attn_mask_start_row_indices.unsqueeze(-1),
                    causal=True,
                )
        else:
            attention_mask = _gen_from_sparse_attn_mask_indices(
                attn_mask_start_row_indices, q.dtype
            )
            if rr_flash_attn is None:
                out = F.scaled_dot_product_attention(
                    q,
                    k,
                    v,
                    attn_mask=attention_mask,
                    is_causal=False,
                )
            else:
                out = rr_flash_attn(
                    F.scaled_dot_product_attention,
                    q,
                    k,
                    v,
                    attn_mask=attention_mask,
                    is_causal=False,
                )
        weights = None
    else:
        if rr_flash_attn is None:
            out = F.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=attention_mask,
                is_causal=attention_mask is None and q.shape[1] != 1,
            )
            weights = None
        else:
            out = rr_flash_attn(
                F.scaled_dot_product_attention,
                q,
                k,
                v,
                attn_mask=attention_mask,
                is_causal=attention_mask is None and q.shape[1] != 1,
            )
            weights = None

    out = tensor.reshape(x=out, shape=[0, 0, out.shape[2] * out.shape[3]])
    return out, weights


def _gen_from_sparse_attn_mask_indices(attn_mask_start_row_indices, dtype):
    """
    Recover 4-D attention_mask from attn_mask_start_row_indices.

    Args:
        attn_mask_start_row_indices (paddle.Tensor): The start row indices for the attention mask.
        dtype (str): The data type of the tensor.

    Returns:
        paddle.Tensor: The dense attention mask recovered from attn_mask_start_row_indices.
    """
    batch_size, _, max_seq_len = attn_mask_start_row_indices.shape
    base = (
        paddle.arange(max_seq_len, dtype="int32")
        .unsqueeze(1)
        .expand([batch_size, -1, max_seq_len])
        .unsqueeze(1)
    )
    mask_indices = attn_mask_start_row_indices.unsqueeze(1)

    tril = paddle.tril(
        paddle.ones([max_seq_len, max_seq_len], dtype="bool").expand(
            [batch_size, 1, max_seq_len, max_seq_len]
        )
    )
    attention_mask = paddle.logical_and(base < mask_indices, tril)
    attention_mask = paddle.scale(
        x=attention_mask.astype(dtype),
        scale=1000000.0,
        bias=-1.0,
        bias_after_scale=False,
    )

    return attention_mask



#!/usr/bin/env python

# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import pprint
import sys

from paddlex_hps_client import triton_request, utils
from tritonclient import grpc as triton_grpc

OUTPUT_IMAGE_PATH = "out.jpg"


def parse_image_label_pairs(image_label_pairs):
    if len(image_label_pairs) % 2 != 0:
        raise ValueError("The number of image-label pairs must be even.")
    return [
        {"image": utils.prepare_input_file(img), "label": lab}
        for img, lab in zip(image_label_pairs[0::2], image_label_pairs[1::2])
    ]


def create_triton_client(url):
    return triton_grpc.InferenceServerClient(url)


def ensure_no_error(output):
    if output["errorCode"] != 0:
        print(f"Error code: {output['errorCode']}", file=sys.stderr)
        print(f"Error message: {output['errorMsg']}", file=sys.stderr)
        sys.exit(1)


def do_index_build(args):
    client = create_triton_client(args.url)
    if args.image_label_pairs:
        image_label_pairs = parse_image_label_pairs(args.image_label_pairs)
    else:
        image_label_pairs = []
    input_ = {"imageLabelPairs": image_label_pairs}
    output = triton_request(client, "face-recognition-index-build", input_)
    ensure_no_error(output)
    result = output["result"]
    pprint.pp(result)


def do_index_add(args):
    client = create_triton_client(args.url)
    image_label_pairs = parse_image_label_pairs(args.image_label_pairs)
    input_ = {"imageLabelPairs": image_label_pairs}
    if args.index_key is not None:
        input_["indexKey"] = args.index_key
    output = triton_request(client, "face-recognition-index-add", input_)
    ensure_no_error(output)
    result = output["result"]
    pprint.pp(result)


def do_index_remove(args):
    client = create_triton_client(args.url)
    input_ = {"ids": args.ids}
    if args.index_key is not None:
        input_["indexKey"] = args.index_key
    output = triton_request(client, "face-recognition-index-remove", input_)
    ensure_no_error(output)
    result = output["result"]
    pprint.pp(result)


def do_infer(args):
    client = create_triton_client(args.url)
    input_ = {"image": utils.prepare_input_file(args.image)}
    if args.index_key is not None:
        input_["indexKey"] = args.index_key
    if args.no_visualization:
        input_["visualize"] = False
    output = triton_request(client, "face-recognition-infer", input_)
    ensure_no_error(output)
    result = output["result"]
    utils.save_output_file(result["image"], OUTPUT_IMAGE_PATH)
    print(f"Output image saved at {OUTPUT_IMAGE_PATH}")
    print("\nDetected faces:")
    pprint.pp(result["faces"])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--url", type=str, default="localhost:8001")

    subparsers = parser.add_subparsers(dest="cmd")

    parser_index_build = subparsers.add_parser("index-build")
    parser_index_build.add_argument("--image-label-pairs", type=str, nargs="+")
    parser_index_build.set_defaults(func=do_index_build)

    parser_index_add = subparsers.add_parser("index-add")
    parser_index_add.add_argument(
        "--image-label-pairs", type=str, nargs="+", required=True
    )
    parser_index_add.add_argument("--index-key", type=str, required=True)
    parser_index_add.set_defaults(func=do_index_add)

    parser_index_remove = subparsers.add_parser("index-remove")
    parser_index_remove.add_argument("--ids", type=int, nargs="+", required=True)
    parser_index_remove.add_argument("--index-key", type=str, required=True)
    parser_index_remove.set_defaults(func=do_index_remove)

    parser_infer = subparsers.add_parser("infer")
    parser_infer.add_argument("--image", type=str, required=True)
    parser_infer.add_argument("--index-key", type=str)
    parser.add_argument("--no-visualization", action="store_true")

    parser_infer.set_defaults(func=do_infer)

    args = parser.parse_args()

    args.func(args)


if __name__ == "__main__":
    main()



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from typing import Union, List
import logging
from .... import UltraInferModel, ModelFormat
from .... import c_lib_wrap as C
from ...common import ProcessorManager


class PaddleDetPreprocessor(ProcessorManager):
    def __init__(self, config_file):
        """Create a preprocessor for PaddleDetection Model from configuration file

        :param config_file: (str)Path of configuration file, e.g ppyoloe/infer_cfg.yml
        """
        self._manager = C.vision.detection.PaddleDetPreprocessor(config_file)

    def disable_normalize(self):
        """
        This function will disable normalize in preprocessing step.
        """
        self._manager.disable_normalize()

    def disable_permute(self):
        """
        This function will disable hwc2chw in preprocessing step.
        """
        self._manager.disable_permute()


class NMSOption:
    def __init__(self):
        self.nms_option = C.vision.detection.NMSOption()

    @property
    def background_label(self):
        return self.nms_option.background_label


class NMSRotatedOption:
    def __init__(self):
        self.nms_rotated_option = C.vision.detection.NMSRotatedOption()

    @property
    def background_label(self):
        return self.nms_rotated_option.background_label


class PaddleDetPostprocessor:
    def __init__(self):
        """Create a postprocessor for PaddleDetection Model"""
        self._postprocessor = C.vision.detection.PaddleDetPostprocessor()

    def run(self, runtime_results):
        """Postprocess the runtime results for PaddleDetection Model

        :param: runtime_results: (list of FDTensor)The output FDTensor results from runtime
        :return: list of ClassifyResult(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results)

    def apply_nms(self):
        self._postprocessor.apply_nms()

    def set_nms_option(self, nms_option=None):
        """This function will enable decode and nms in postprocess step."""
        if nms_option is None:
            nms_option = NMSOption()
        self._postprocessor.set_nms_option(self, nms_option.nms_option)

    def set_nms_rotated_option(self, nms_rotated_option=None):
        """This function will enable decode and rotated nms in postprocess step."""
        if nms_rotated_option is None:
            nms_rotated_option = NMSRotatedOption()
        self._postprocessor.set_nms_rotated_option(
            self, nms_rotated_option.nms_rotated_option
        )


class PPYOLOE(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a PPYOLOE model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g ppyoloe/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g ppyoloe/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """
        super(PPYOLOE, self).__init__(runtime_option)

        self._model = C.vision.detection.PPYOLOE(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PPYOLOE model initialize failed."

    def predict(self, im):
        """Detect an input image

        :param im: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: DetectionResult
        """

        assert im is not None, "The input image data is None."
        return self._model.predict(im)

    def batch_predict(self, images):
        """Detect a batch of input image list

        :param im: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return list of DetectionResult
        """

        return self._model.batch_predict(images)

    def clone(self):
        """Clone PPYOLOE object

        :return: a new PPYOLOE object
        """

        class PPYOLOEClone(PPYOLOE):
            def __init__(self, model):
                self._model = model

        clone_model = PPYOLOEClone(self._model.clone())
        return clone_model

    @property
    def preprocessor(self):
        """Get PaddleDetPreprocessor object of the loaded model

        :return PaddleDetPreprocessor
        """
        return self._model.preprocessor

    @property
    def postprocessor(self):
        """Get PaddleDetPostprocessor object of the loaded model

        :return PaddleDetPostprocessor
        """
        return self._model.postprocessor


class PPYOLO(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a PPYOLO model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g ppyolo/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g ppyolo/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "PPYOLO model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.PPYOLO(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PPYOLO model initialize failed."

    def clone(self):
        """Clone PPYOLO object

        :return: a new PPYOLO object
        """

        class PPYOLOClone(PPYOLO):
            def __init__(self, model):
                self._model = model

        clone_model = PPYOLOClone(self._model.clone())
        return clone_model


class PaddleYOLOX(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a YOLOX model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g yolox/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g yolox/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "PaddleYOLOX model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.PaddleYOLOX(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PaddleYOLOX model initialize failed."

    def clone(self):
        """Clone PaddleYOLOX object

        :return: a new PaddleYOLOX object
        """

        class PaddleYOLOXClone(PaddleYOLOX):
            def __init__(self, model):
                self._model = model

        clone_model = PaddleYOLOXClone(self._model.clone())
        return clone_model


class PicoDet(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a PicoDet model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g picodet/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g picodet/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        self._model = C.vision.detection.PicoDet(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PicoDet model initialize failed."

    def clone(self):
        """Clone PicoDet object

        :return: a new PicoDet object
        """

        class PicoDetClone(PicoDet):
            def __init__(self, model):
                self._model = model

        clone_model = PicoDetClone(self._model.clone())
        return clone_model


class FasterRCNN(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a FasterRCNN model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g fasterrcnn/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g fasterrcnn/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "FasterRCNN model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.FasterRCNN(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "FasterRCNN model initialize failed."

    def clone(self):
        """Clone FasterRCNN object

        :return: a new FasterRCNN object
        """

        class FasterRCNNClone(FasterRCNN):
            def __init__(self, model):
                self._model = model

        clone_model = FasterRCNNClone(self._model.clone())
        return clone_model


class YOLOv3(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a YOLOv3 model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g yolov3/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g yolov3/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "YOLOv3 model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.YOLOv3(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "YOLOv3 model initialize failed."

    def clone(self):
        """Clone YOLOv3 object

        :return: a new YOLOv3 object
        """

        class YOLOv3Clone(YOLOv3):
            def __init__(self, model):
                self._model = model

        clone_model = YOLOv3Clone(self._model.clone())
        return clone_model


class SOLOv2(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a SOLOv2 model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g solov2/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g solov2/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g solov2/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "SOLOv2 model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.SOLOv2(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "SOLOv2 model initialize failed."

    def clone(self):
        """Clone SOLOv2 object

        :return: a new SOLOv2 object
        """

        class SOLOv2Clone(SOLOv2):
            def __init__(self, model):
                self._model = model

        clone_model = SOLOv2Clone(self._model.clone())
        return clone_model


class MaskRCNN(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a MaskRCNN model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g fasterrcnn/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g fasterrcnn/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "MaskRCNN model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.MaskRCNN(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "MaskRCNN model initialize failed."

    def batch_predict(self, images):
        """Detect a batch of input image list, batch_predict is not supported for maskrcnn now.

        :param im: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return list of DetectionResult
        """

        raise Exception("batch_predict is not supported for MaskRCNN model now.")

    def clone(self):
        """Clone MaskRCNN object

        :return: a new MaskRCNN object
        """

        class MaskRCNNClone(MaskRCNN):
            def __init__(self, model):
                self._model = model

        clone_model = MaskRCNNClone(self._model.clone())
        return clone_model


class SSD(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a SSD model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g ssd/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g ssd/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "SSD model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.SSD(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "SSD model initialize failed."

    def clone(self):
        """Clone SSD object

        :return: a new SSD object
        """

        class SSDClone(SSD):
            def __init__(self, model):
                self._model = model

        clone_model = SSDClone(self._model.clone())
        return clone_model


class PaddleYOLOv5(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a YOLOv5 model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g yolov5/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g yolov5/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "PaddleYOLOv5 model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.PaddleYOLOv5(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PaddleYOLOv5 model initialize failed."


class PaddleYOLOv6(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a YOLOv6 model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g yolov6/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g yolov6/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "PaddleYOLOv6 model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.PaddleYOLOv6(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PaddleYOLOv6 model initialize failed."


class PaddleYOLOv7(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a YOLOv7 model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g yolov7/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g yolov7/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "PaddleYOLOv7 model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.PaddleYOLOv7(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PaddleYOLOv7 model initialize failed."


class PaddleYOLOv8(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a YOLOv8 model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g yolov8/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g yolov8/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g yolov8/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        self._model = C.vision.detection.PaddleYOLOv8(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PaddleYOLOv8 model initialize failed."


class RTMDet(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a RTMDet model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g rtmdet/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g rtmdet/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "RTMDet model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.RTMDet(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "RTMDet model initialize failed."


class CascadeRCNN(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a CascadeRCNN model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g cascadercnn/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g cascadercnn/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "CascadeRCNN model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.CascadeRCNN(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "CascadeRCNN model initialize failed."


class PSSDet(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a PSSDet model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g pssdet/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g pssdet/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "PSSDet model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.PSSDet(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PSSDet model initialize failed."


class RetinaNet(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a RetinaNet model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g retinanet/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g retinanet/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "RetinaNet model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.RetinaNet(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "RetinaNet model initialize failed."


class PPYOLOESOD(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a PPYOLOESOD model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g ppyoloesod/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g ppyoloesod/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "PPYOLOESOD model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.PPYOLOESOD(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PPYOLOESOD model initialize failed."


class FCOS(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a FCOS model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g fcos/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g fcos/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "FCOS model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.FCOS(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "FCOS model initialize failed."


class TTFNet(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a TTFNet model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g ttfnet/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g ttfnet/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "TTFNet model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.TTFNet(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "TTFNet model initialize failed."


class TOOD(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a TOOD model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g tood/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g tood/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "TOOD model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.TOOD(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "TOOD model initialize failed."


class GFL(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a GFL model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g gfl/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g gfl/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        assert (
            model_format == ModelFormat.PADDLE
        ), "GFL model only support model format of ModelFormat.Paddle now."
        self._model = C.vision.detection.GFL(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "GFL model initialize failed."


class PaddleDetectionModel(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a PaddleDetectionModel model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g ppyoloe/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g ppyoloe/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """
        super(PaddleDetectionModel, self).__init__(runtime_option)

        self._model = C.vision.detection.PaddleDetectionModel(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PaddleDetectionModel model initialize failed."

    def predict(self, im):
        """Detect an input image

        :param im: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: DetectionResult
        """

        assert im is not None, "The input image data is None."
        return self._model.predict(im)

    def batch_predict(self, images):
        """Detect a batch of input image list

        :param im: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return list of DetectionResult
        """

        return self._model.batch_predict(images)

    def clone(self):
        """Clone PPYOLOE object

        :return: a new PPYOLOE object
        """

        class PPYOLOEClone(PPYOLOE):
            def __init__(self, model):
                self._model = model

        clone_model = PPYOLOEClone(self._model.clone())
        return clone_model

    @property
    def preprocessor(self):
        """Get PaddleDetPreprocessor object of the loaded model

        :return PaddleDetPreprocessor
        """
        return self._model.preprocessor

    @property
    def postprocessor(self):
        """Get PaddleDetPostprocessor object of the loaded model

        :return PaddleDetPostprocessor
        """
        return self._model.postprocessor


class PPYOLOER(PPYOLOE):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a PPYOLOER model exported by PaddleDetection.

        :param model_file: (str)Path of model file, e.g ppyoloe_r/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g ppyoloe_r/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str)Path of configuration file for deployment, e.g ppyoloe_r/infer_cfg.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """

        super(PPYOLOE, self).__init__(runtime_option)

        self._model = C.vision.detection.PPYOLOER(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PicoDet model initialize failed."

    def clone(self):
        """Clone PPYOLOER object

        :return: a new PPYOLOER object
        """

        class PPYOLOERClone(PPYOLOER):
            def __init__(self, model):
                self._model = model

        clone_model = PPYOLOERClone(self._model.clone())
        return clone_model



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import cv2
import numpy as np


class Resize(object):
    def __init__(self, size=(640, 640), **kwargs):
        self.size = size

    def resize_image(self, img):
        resize_h, resize_w = self.size
        ori_h, ori_w = img.shape[:2]  # (h, w, c)
        ratio_h = float(resize_h) / ori_h
        ratio_w = float(resize_w) / ori_w
        img = cv2.resize(img, (int(resize_w), int(resize_h)))
        return img, [ratio_h, ratio_w]

    def __call__(self, data):
        img = data["image"]
        if "polys" in data:
            text_polys = data["polys"]

        img_resize, [ratio_h, ratio_w] = self.resize_image(img)
        if "polys" in data:
            new_boxes = []
            for box in text_polys:
                new_box = []
                for cord in box:
                    new_box.append([cord[0] * ratio_w, cord[1] * ratio_h])
                new_boxes.append(new_box)
            data["polys"] = np.array(new_boxes, dtype=np.float32)
        data["image"] = img_resize
        return data


class NormalizeImage(object):
    """normalize image such as subtract mean, divide std"""

    def __init__(self, scale=None, mean=None, std=None, order="chw", **kwargs):
        if isinstance(scale, str):
            scale = eval(scale)
        self.scale = np.float32(scale if scale is not None else 1.0 / 255.0)
        mean = mean if mean is not None else [0.485, 0.456, 0.406]
        std = std if std is not None else [0.229, 0.224, 0.225]

        shape = (3, 1, 1) if order == "chw" else (1, 1, 3)
        self.mean = np.array(mean).reshape(shape).astype("float32")
        self.std = np.array(std).reshape(shape).astype("float32")

    def __call__(self, data):
        img = data["image"]
        from PIL import Image

        if isinstance(img, Image.Image):
            img = np.array(img)
        assert isinstance(img, np.ndarray), "invalid input 'img' in NormalizeImage"
        data["image"] = (img.astype("float32") * self.scale - self.mean) / self.std
        return data


class ToCHWImage(object):
    """convert hwc image to chw image"""

    def __init__(self, **kwargs):
        pass

    def __call__(self, data):
        img = data["image"]
        from PIL import Image

        if isinstance(img, Image.Image):
            img = np.array(img)
        data["image"] = img.transpose((2, 0, 1))
        return data


class KeepKeys(object):
    def __init__(self, keep_keys, **kwargs):
        self.keep_keys = keep_keys

    def __call__(self, data):
        data_list = []
        for key in self.keep_keys:
            data_list.append(data[key])
        return data_list



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import logging
import os
import sys
import platform

# Create a symbol link to tensorrt library.
trt_directory = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "libs/third_libs/tensorrt/lib/"
)
if os.name != "nt" and os.path.exists(trt_directory):
    logging.basicConfig(level=logging.INFO)
    for trt_lib in [
        "libnvcaffe_parser.so",
        "libnvinfer_plugin.so",
        "libnvinfer.so",
        "libnvonnxparser.so",
        "libnvparsers.so",
    ]:
        dst = os.path.join(trt_directory, trt_lib)
        src = os.path.join(trt_directory, trt_lib + ".8")
        if not os.path.exists(dst):
            try:
                os.symlink(src, dst)
                logging.info(f"Create a symbolic link pointing to {src} named {dst}.")
            except OSError as e:
                logging.warning(
                    f"Failed to create a symbolic link pointing to {src} by an unprivileged user. "
                    "It may failed when you use Paddle TensorRT backend. "
                    "Please use administrator privilege to import ultra_infer at first time."
                )
                break

    # HACK: Reset the root logger config that got messed up by FD.
    root_logger = logging.getLogger()
    root_logger.level = logging.WARNING
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

from .code_version import version, git_version, extra_version_info
from .code_version import enable_trt_backend, enable_paddle_backend, with_gpu

# Note(zhoushunjie): Fix the import order of paddle and ultra_infer library.
# This solution will be removed it when the conflict of paddle and
# ultra_infer is fixed.

# Note(qiuyanjun): Add backward compatible for paddle 2.4.x
sys_platform = platform.platform().lower()


def get_paddle_version():
    paddle_version = ""
    try:
        import pkg_resources

        paddle_version = pkg_resources.require("paddlepaddle-gpu")[0].version.split(
            ".post"
        )[0]
    except:
        try:
            paddle_version = pkg_resources.require("paddlepaddle")[0].version.split(
                ".post"
            )[0]
        except:
            pass
    return paddle_version


def should_import_paddle():
    if ("paddle2.4" in extra_version_info) or ("post24" in extra_version_info):
        paddle_version = get_paddle_version()
        if (
            paddle_version != ""
            and paddle_version <= "2.4.2"
            and paddle_version != "0.0.0"
        ):
            return True
    return False


def should_set_tensorrt():
    if (
        with_gpu == "ON"
        and enable_paddle_backend == "ON"
        and enable_trt_backend == "ON"
    ):
        return True
    return False


def tensorrt_is_avaliable():
    # Note(qiuyanjun): Only support linux now.
    found_trt_lib = False
    if ("linux" in sys_platform) and ("LD_LIBRARY_PATH" in os.environ.keys()):
        for lib_path in os.environ["LD_LIBRARY_PATH"].split(":"):
            if os.path.exists(os.path.join(lib_path, "libnvinfer.so")):
                found_trt_lib = True
                break
    return found_trt_lib


try:
    # windows: no conflict between ultra_infer and paddle.
    # linux: must import paddle first to solve the conflict.
    # macos: still can not solve the conflict between ultra_infer and paddle,
    #        due to the global flags redefined in paddle/paddle_inference so.
    #        we got the error (ERROR: flag 'xxx' was defined more than once).
    if "linux" in sys_platform:
        if should_import_paddle():
            import paddle  # need import paddle first for paddle2.4.x

            # check whether tensorrt in LD_LIBRARY_PATH for ultra_infer
            if should_set_tensorrt() and (not tensorrt_is_avaliable()):
                if os.path.exists(trt_directory):
                    logging.info(
                        "\n[WARNING] Can not find TensorRT lib in LD_LIBRARY_PATH for UltraInfer! \
            \n[WARNING] Please export [ YOUR CUSTOM TensorRT ] lib path to LD_LIBRARY_PATH first, or run the command: \
            \n[WARNING] Linux: 'export LD_LIBRARY_PATH=$(python -c 'from ultra_infer import trt_directory; print(trt_directory)'):$LD_LIBRARY_PATH'"
                    )
                else:
                    logging.info(
                        "\n[WARNING] Can not find TensorRT lib in LD_LIBRARY_PATH for UltraInfer! \
            \n[WARNING] Please export [YOUR CUSTOM TensorRT] lib path to LD_LIBRARY_PATH first."
                    )
except:
    pass


from .c_lib_wrap import (
    ModelFormat,
    Backend,
    FDDataType,
    TensorInfo,
    Device,
    is_built_with_gpu,
    is_built_with_ort,
    ModelFormat,
    is_built_with_paddle,
    is_built_with_trt,
    get_default_cuda_directory,
    is_built_with_openvino,
    is_built_with_om,
)


def set_logger(enable_info=True, enable_warning=True):
    """Set behaviour of logger while using UltraInfer

    :param enable_info: (boolean)Whether to print out log level of INFO
    :param enable_warning: (boolean)Whether to print out log level of WARNING, recommend to set to True
    """
    from .c_lib_wrap import set_logger

    set_logger(enable_info, enable_warning)


from .runtime import Runtime, RuntimeOption
from .model import UltraInferModel
from . import c_lib_wrap as C
from . import vision
from . import pipeline
from . import text
from . import ts
from .download import download, download_and_decompress, download_model, get_model_list


__version__ = version



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Tuple, Union

import paddle
import paddle.nn as nn
import paddle.nn.functional as F
from paddle import ParamAttr
from paddle.nn.initializer import Constant, KaimingNormal
from paddle.regularizer import L2Decay

from ....utils.benchmark import add_inference_operations, benchmark
from ...common.transformers.transformers import (
    BatchNormHFStateDictMixin,
    PretrainedModel,
)
from ._config_pp_ocrv5_server import PPOCRV5ServerDetConfig
from .pp_ocrv5_modules import DBHead, LearnableAffineBlock

kaiming_normal_ = KaimingNormal()
zeros_ = Constant(value=0.0)
ones_ = Constant(value=1.0)


class ConvBNAct(nn.Layer):
    """
    ConvBNAct: Convolution + Batch Normalization + Activation (optional) with Learnable Affine Block

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int, optional): Convolution kernel size, default is 3
        stride (int, optional): Convolution stride, default is 1
        padding (Union[int, str], optional): Padding value or mode (e.g. 'same'), default is 1
        groups (int, optional): Number of grouped convolution groups, default is 1
        use_act (bool, optional): Whether to apply ReLU activation, default is True
        use_lab (bool, optional): Whether to use LearnableAffineBlock after activation, default is False
        lr_mult (float, optional): Learning rate multiplier for conv/bn parameters, default is 1.0

    Returns:
        paddle.Tensor: Output tensor after convolution, BN and optional activation/affine transform
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: Union[int, str] = 1,
        groups: int = 1,
        use_act: bool = True,
        use_lab: bool = False,
        lr_mult: float = 1.0,
    ) -> None:
        super().__init__()
        self.use_act = use_act
        self.use_lab = use_lab
        padding_val = padding if isinstance(padding, str) else (kernel_size - 1) // 2
        self.conv = nn.Conv2D(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding=padding_val,
            groups=groups,
            weight_attr=ParamAttr(learning_rate=lr_mult),
            bias_attr=False,
        )
        self.bn = nn.BatchNorm2D(
            out_channels,
            weight_attr=ParamAttr(regularizer=L2Decay(0.0), learning_rate=lr_mult),
            bias_attr=ParamAttr(regularizer=L2Decay(0.0), learning_rate=lr_mult),
        )
        if self.use_act:
            self.act = nn.ReLU()
            if self.use_lab:
                self.lab = LearnableAffineBlock(lr_mult=lr_mult)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.conv(x)
        x = self.bn(x)
        if self.use_act:
            x = self.act(x)
            if self.use_lab:
                x = self.lab(x)
        return x


class LightConvBNAct(nn.Layer):
    """
    LightConvBNAct: Lightweight depthwise separable convolution block with BN and activation

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int): Kernel size of depthwise convolution
        use_lab (bool, optional): Whether to use LearnableAffineBlock in ConvBNAct, default is False
        lr_mult (float, optional): Learning rate multiplier for conv/bn parameters, default is 1.0
        **kwargs: Additional keyword arguments

    Returns:
        paddle.Tensor: Output tensor after pointwise conv (no act) + depthwise conv (with act)
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        use_lab: bool = False,
        lr_mult: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__()
        self.conv1 = ConvBNAct(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=1,
            use_act=False,
            use_lab=use_lab,
            lr_mult=lr_mult,
        )
        self.conv2 = ConvBNAct(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            groups=out_channels,
            use_act=True,
            use_lab=use_lab,
            lr_mult=lr_mult,
        )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        return x


class StemBlock(nn.Layer):
    """
    StemBlock: Multi-stage convolution stem block with pooling and concatenation

    Args:
        in_channels (int): Number of input channels
        mid_channels (int): Number of intermediate channels for stem layers
        out_channels (int): Number of output channels
        use_lab (bool, optional): Whether to use LearnableAffineBlock in ConvBNAct, default is False
        lr_mult (float, optional): Learning rate multiplier for conv/bn parameters, default is 1.0

    Returns:
        paddle.Tensor: Output tensor after multi-stage convolution, pooling and concatenation
    """

    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        use_lab: bool = False,
        lr_mult: float = 1.0,
    ) -> None:
        super().__init__()
        self.stem1 = ConvBNAct(
            in_channels=in_channels,
            out_channels=mid_channels,
            kernel_size=3,
            stride=2,
            use_lab=use_lab,
            lr_mult=lr_mult,
        )
        self.stem2a = ConvBNAct(
            in_channels=mid_channels,
            out_channels=mid_channels // 2,
            kernel_size=2,
            stride=1,
            padding="SAME",
            use_lab=use_lab,
            lr_mult=lr_mult,
        )
        self.stem2b = ConvBNAct(
            in_channels=mid_channels // 2,
            out_channels=mid_channels,
            kernel_size=2,
            stride=1,
            padding="SAME",
            use_lab=use_lab,
            lr_mult=lr_mult,
        )
        self.stem3 = ConvBNAct(
            in_channels=mid_channels * 2,
            out_channels=mid_channels,
            kernel_size=3,
            stride=2,
            use_lab=use_lab,
            lr_mult=lr_mult,
        )
        self.stem4 = ConvBNAct(
            in_channels=mid_channels,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            use_lab=use_lab,
            lr_mult=lr_mult,
        )
        self.pool = nn.MaxPool2D(
            kernel_size=2, stride=1, ceil_mode=True, padding="SAME"
        )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.stem1(x)
        x2 = self.stem2a(x)
        x2 = self.stem2b(x2)
        x1 = self.pool(x)
        x = paddle.concat([x1, x2], 1)
        x = self.stem3(x)
        x = self.stem4(x)

        return x


class HGV2_Block(nn.Layer):
    """
    HGV2_Block: Multi-layer convolution block with feature aggregation and residual connection

    Args:
        in_channels (int): Number of input channels
        mid_channels (int): Number of intermediate channels for each layer
        out_channels (int): Number of output channels after aggregation
        kernel_size (int, optional): Kernel size of convolution layers, default is 3
        layer_num (int, optional): Number of convolution layers in the block, default is 6
        identity (bool, optional): Whether to add identity residual connection, default is False
        light_block (bool, optional): Whether to use LightConvBNAct (True) or ConvBNAct (False), default is True
        use_lab (bool, optional): Whether to use LearnableAffineBlock in conv layers, default is False
        lr_mult (float, optional): Learning rate multiplier for conv/bn parameters, default is 1.0

    Returns:
        paddle.Tensor: Output tensor after multi-layer conv, feature concatenation and aggregation
    """

    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        layer_num: int = 6,
        identity: bool = False,
        light_block: bool = True,
        use_lab: bool = False,
        lr_mult: float = 1.0,
    ) -> None:
        super().__init__()
        self.identity = identity

        self.layers = nn.LayerList()
        block_type = "LightConvBNAct" if light_block else "ConvBNAct"
        for i in range(layer_num):
            self.layers.append(
                eval(block_type)(
                    in_channels=in_channels if i == 0 else mid_channels,
                    out_channels=mid_channels,
                    stride=1,
                    kernel_size=kernel_size,
                    use_lab=use_lab,
                    lr_mult=lr_mult,
                )
            )
        # feature aggregation
        total_channels = in_channels + layer_num * mid_channels
        self.aggregation_squeeze_conv = ConvBNAct(
            in_channels=total_channels,
            out_channels=out_channels // 2,
            kernel_size=1,
            stride=1,
            use_lab=use_lab,
            lr_mult=lr_mult,
        )
        self.aggregation_excitation_conv = ConvBNAct(
            in_channels=out_channels // 2,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            use_lab=use_lab,
            lr_mult=lr_mult,
        )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        identity = x
        output = []
        output.append(x)
        for layer in self.layers:
            x = layer(x)
            output.append(x)
        x = paddle.concat(output, axis=1)
        x = self.aggregation_squeeze_conv(x)
        x = self.aggregation_excitation_conv(x)
        if self.identity:
            x += identity
        return x


class HGV2_Stage(nn.Layer):
    """
    HGV2_Stage: Sequential HGV2_Block layers with optional depthwise downsampling

    Args:
        in_channels (int): Number of input channels
        mid_channels (int): Number of intermediate channels for HGV2_Block layers
        out_channels (int): Number of output channels for each HGV2_Block
        block_num (int): Number of HGV2_Block in the stage
        layer_num (int, optional): Number of convolution layers in each HGV2_Block, default is 6
        is_downsample (bool, optional): Whether to apply depthwise downsampling at stage start, default is True
        light_block (bool, optional): Whether to use LightConvBNAct in HGV2_Block, default is True
        kernel_size (int, optional): Kernel size of convolution layers in HGV2_Block, default is 3
        use_lab (bool, optional): Whether to use LearnableAffineBlock in conv layers, default is False
        stride (int, optional): Stride for downsampling convolution, default is 2
        lr_mult (float, optional): Learning rate multiplier for conv/bn parameters, default is 1.0

    Returns:
        paddle.Tensor: Output tensor after optional downsampling and sequential HGV2_Block processing
    """

    def __init__(
        self,
        in_channels: int,
        mid_channels: int,
        out_channels: int,
        block_num: int,
        layer_num: int = 6,
        is_downsample: bool = True,
        light_block: bool = True,
        kernel_size: int = 3,
        use_lab: bool = False,
        stride: int = 2,
        lr_mult: float = 1.0,
    ) -> None:

        super().__init__()
        self.is_downsample = is_downsample
        if self.is_downsample:
            self.downsample = ConvBNAct(
                in_channels=in_channels,
                out_channels=in_channels,
                kernel_size=3,
                stride=stride,
                groups=in_channels,
                use_act=False,
                use_lab=use_lab,
                lr_mult=lr_mult,
            )

        blocks_list = []
        for i in range(block_num):
            blocks_list.append(
                HGV2_Block(
                    in_channels=in_channels if i == 0 else out_channels,
                    mid_channels=mid_channels,
                    out_channels=out_channels,
                    kernel_size=kernel_size,
                    layer_num=layer_num,
                    identity=False if i == 0 else True,
                    light_block=light_block,
                    use_lab=use_lab,
                    lr_mult=lr_mult,
                )
            )
        self.blocks = nn.Sequential(*blocks_list)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        if self.is_downsample:
            x = self.downsample(x)
        x = self.blocks(x)
        return x


class PPHGNetV2(nn.Layer):
    """
    PPHGNetV2: Hierarchical feature extraction network with stem block and multi-stage HGV2 blocks

    Args:
        stage_config (Dict[str, Tuple]): Dictionary of stage configurations, each tuple contains (in_channels, mid_channels, out_channels, block_num, is_downsample, light_block, kernel_size, layer_num, stride)
        stem_channels (Tuple[int, int, int]): Stem block channels (in, mid, out)
        use_lab (bool): Whether to use LearnableAffineBlock in ConvBNAct layers
        use_last_conv (bool): Whether to use last convolution layer (unused in current implementation)
        class_expand (float): Expansion factor for classification head channels
        class_num (int): Number of classification classes
        lr_mult_list (List[float]): Learning rate multipliers for stem and each stage
        det (bool): Whether the network is used for detection (controls output indices)
        out_indices (List[int]): Indices of stages to output features for detection
        **kwargs: Additional keyword arguments

    Returns:
        List[paddle.Tensor]: List of feature tensors from specified stages (only when det=True)
    """

    def __init__(
        self,
        stage_config: Dict[str, Tuple],
        stem_channels: Tuple[int, int, int],
        use_lab: bool,
        use_last_conv: bool,
        class_expand: float,
        class_num: int,
        lr_mult_list: List[float],
        det: bool,
        out_indices: List[int],
        **kwargs,
    ) -> None:
        super().__init__()
        self.det = det
        self.use_lab = use_lab
        self.use_last_conv = use_last_conv
        self.class_expand = class_expand
        self.class_num = class_num
        self.out_indices = out_indices
        self.out_channels = []

        # stem
        self.stem = StemBlock(
            in_channels=stem_channels[0],
            mid_channels=stem_channels[1],
            out_channels=stem_channels[2],
            use_lab=use_lab,
            lr_mult=lr_mult_list[0],
        )

        # stages
        self.stages = nn.LayerList()
        for i, k in enumerate(stage_config):
            (
                in_channels,
                mid_channels,
                out_channels,
                block_num,
                is_downsample,
                light_block,
                kernel_size,
                layer_num,
                stride,
            ) = stage_config[k]
            self.stages.append(
                HGV2_Stage(
                    in_channels,
                    mid_channels,
                    out_channels,
                    block_num,
                    layer_num,
                    is_downsample,
                    light_block,
                    kernel_size,
                    use_lab,
                    stride,
                    lr_mult=lr_mult_list[i + 1],
                )
            )
            if i in self.out_indices:
                self.out_channels.append(out_channels)

        self.avg_pool = nn.AdaptiveAvgPool2D(1)

        self._init_weights()

    def _init_weights(self) -> None:
        for m in self.sublayers():
            if isinstance(m, nn.Conv2D):
                kaiming_normal_(m.weight)
            elif isinstance(m, (nn.BatchNorm2D)):
                ones_(m.weight)
                zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                zeros_(m.bias)

    def forward(self, x: paddle.Tensor) -> List[paddle.Tensor]:
        x = self.stem(x)
        out = []
        for i, stage in enumerate(self.stages):
            x = stage(x)
            if self.det and i in self.out_indices:
                out.append(x)
        return out


class DSConv(nn.Layer):
    """
    DSConv: Depthwise separable convolution with bottleneck and residual connection

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int): Kernel size of depthwise convolution
        padding (Union[int, str]): Padding value or mode (e.g. 'SAME') for depthwise conv
        stride (int, optional): Stride for depthwise convolution, default is 1
        groups (Optional[int], optional): Number of groups for depthwise conv (default: in_channels)
        if_act (bool, optional): Whether to apply activation after second BN, default is True
        act (str, optional): Activation function type ('relu' or 'hardswish'), default is 'relu'
        **kwargs: Additional keyword arguments

    Returns:
        paddle.Tensor: Output tensor after depthwise separable convolution and optional residual connection
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        padding: Union[int, str],
        stride: int = 1,
        groups: Optional[int] = None,
        if_act: bool = True,
        act: str = "relu",
        **kwargs,
    ) -> None:
        super(DSConv, self).__init__()
        if groups == None:
            groups = in_channels
        self.if_act = if_act
        self.act = act
        self.conv1 = nn.Conv2D(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=groups,
            bias_attr=False,
        )

        self.bn1 = nn.BatchNorm(num_channels=in_channels, act=None)

        self.conv2 = nn.Conv2D(
            in_channels=in_channels,
            out_channels=int(in_channels * 4),
            kernel_size=1,
            stride=1,
            bias_attr=False,
        )

        self.bn2 = nn.BatchNorm(num_channels=int(in_channels * 4), act=None)

        self.conv3 = nn.Conv2D(
            in_channels=int(in_channels * 4),
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            bias_attr=False,
        )
        self._c = [in_channels, out_channels]
        if in_channels != out_channels:
            self.conv_end = nn.Conv2D(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=1,
                bias_attr=False,
            )

    def forward(self, inputs: paddle.Tensor) -> paddle.Tensor:
        x = self.conv1(inputs)
        x = self.bn1(x)

        x = self.conv2(x)
        x = self.bn2(x)
        if self.if_act:
            if self.act == "relu":
                x = F.relu(x)
            elif self.act == "hardswish":
                x = F.hardswish(x)
            else:
                print(
                    "The activation function({}) is selected incorrectly.".format(
                        self.act
                    )
                )
                exit()

        x = self.conv3(x)
        if self._c[0] != self._c[1]:
            x = x + self.conv_end(inputs)
        return x


class IntraCLBlock(nn.Layer):
    """
    IntraCLBlock: Multi-scale convolution block with vertical/horizontal kernel fusion

    Args:
        in_channels (int): Number of input channels
        reduce_factor (int): Channel reduction ratio for 1x1 convolution
        intraclblock_config (dict): Configuration dict for convolution layers, includes:
            - reduce_channel: (kernel_size, stride, padding) for channel reduction 1x1 conv
            - return_channel: (kernel_size, stride, padding) for channel recovery 1x1 conv
            - v_layer_7x1/5x1/3x1: (kernel_size, stride, padding) for vertical (Hx1) conv
            - q_layer_1x7/1x5/1x3: (kernel_size, stride, padding) for horizontal (1xW) conv
            - c_layer_7x7/5x5/3x3: (kernel_size, stride, padding) for cross (HxW) conv

    Returns:
        paddle.Tensor: Output tensor after multi-scale conv fusion and residual connection
    """

    def __init__(
        self, in_channels: int, reduce_factor: int, intraclblock_config: dict
    ) -> None:
        super(IntraCLBlock, self).__init__()

        self.channels = in_channels
        self.reduce_factor = reduce_factor
        self.intraclblock_config = intraclblock_config

        reduced_ch = self.channels // self.reduce_factor

        self.conv1x1_reduce_channel = nn.Conv2d(
            self.channels, reduced_ch, *self.intraclblock_config["reduce_channel"]
        )
        self.conv1x1_return_channel = nn.Conv2d(
            reduced_ch, self.channels, *self.intraclblock_config["return_channel"]
        )

        self.v_layer_7x1 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["v_layer_7x1"]
        )
        self.v_layer_5x1 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["v_layer_5x1"]
        )
        self.v_layer_3x1 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["v_layer_3x1"]
        )

        self.q_layer_1x7 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["q_layer_1x7"]
        )
        self.q_layer_1x5 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["q_layer_1x5"]
        )
        self.q_layer_1x3 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["q_layer_1x3"]
        )

        self.c_layer_7x7 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["c_layer_7x7"]
        )
        self.c_layer_5x5 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["c_layer_5x5"]
        )
        self.c_layer_3x3 = nn.Conv2d(
            reduced_ch, reduced_ch, *self.intraclblock_config["c_layer_3x3"]
        )

        self.bn = nn.BatchNorm2D(self.channels)
        self.relu = nn.ReLU()

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x_new = self.conv1x1_reduce_channel(x)

        x_7 = (
            self.c_layer_7x7(x_new) + self.v_layer_7x1(x_new) + self.q_layer_1x7(x_new)
        )
        x_5 = self.c_layer_5x5(x_7) + self.v_layer_5x1(x_7) + self.q_layer_1x5(x_7)
        x_3 = self.c_layer_3x3(x_5) + self.v_layer_3x1(x_5) + self.q_layer_1x3(x_5)

        x_relation = self.conv1x1_return_channel(x_3)
        x_relation = self.bn(x_relation)
        x_relation = self.relu(x_relation)

        return x + x_relation


class LKPAN(nn.Layer):
    """
    LKPAN: Feature pyramid network with multi-scale aggregation and IntraCL enhancement

    Args:
        in_channels (List[int]): List of input channel numbers for multi-scale feature maps
        out_channels (int): Number of output channels for 1x1 convolution layers
        mode (str): Network mode ('lite' for DSConv, 'large' for standard Conv2D)
        reduce_factor (int): Channel reduction ratio for IntraCLBlock modules
        intraclblock_config (dict): Configuration dict for convolution layers, includes:
            - reduce_channel: (kernel_size, stride, padding) for channel reduction 1x1 conv
            - return_channel: (kernel_size, stride, padding) for channel recovery 1x1 conv
            - v_layer_7x1/5x1/3x1: (kernel_size, stride, padding) for vertical (Hx1) conv
            - q_layer_1x7/1x5/1x3: (kernel_size, stride, padding) for horizontal (1xW) conv
            - c_layer_7x7/5x5/3x3: (kernel_size, stride, padding) for cross (HxW) conv
        upsample_mode (str): Interpolation mode for upsample operation
        upsample_align_mode (int): Align mode for upsample operation
        **kwargs: Additional keyword arguments

    Returns:
        paddle.Tensor: Fused feature tensor after multi-scale feature aggregation, IntraCLBlock enhancement and concatenation
    """

    def __init__(
        self,
        in_channels: List[int],
        out_channels: int,
        mode: str,
        reduce_factor: int,
        intraclblock_config: dict,
        upsample_mode: str,
        upsample_align_mode: int,
        **kwargs,
    ) -> None:
        super(LKPAN, self).__init__()
        self.out_channels = out_channels
        self.upsample_mode = upsample_mode
        self.upsample_align_mode = upsample_align_mode
        weight_attr = nn.initializer.KaimingUniform()

        if mode.lower() == "lite":
            p_layer = DSConv
        elif mode.lower() == "large":
            p_layer = nn.Conv2D
        else:
            raise ValueError(
                "mode can only be one of ['lite', 'large'], but received {}".format(
                    mode
                )
            )

        self.ins_conv = nn.LayerList()
        self.inp_conv = nn.LayerList()
        self.pan_head_conv = nn.LayerList()
        self.pan_lat_conv = nn.LayerList()

        for i in range(len(in_channels)):
            self.ins_conv.append(
                nn.Conv2D(
                    in_channels=in_channels[i],
                    out_channels=self.out_channels,
                    kernel_size=1,
                    weight_attr=ParamAttr(initializer=weight_attr),
                    bias_attr=False,
                )
            )

            self.inp_conv.append(
                p_layer(
                    in_channels=self.out_channels,
                    out_channels=self.out_channels // 4,
                    kernel_size=9,
                    padding=4,
                    weight_attr=ParamAttr(initializer=weight_attr),
                    bias_attr=False,
                )
            )

            if i > 0:
                self.pan_head_conv.append(
                    nn.Conv2D(
                        in_channels=self.out_channels // 4,
                        out_channels=self.out_channels // 4,
                        kernel_size=3,
                        padding=1,
                        stride=2,
                        weight_attr=ParamAttr(initializer=weight_attr),
                        bias_attr=False,
                    )
                )
            self.pan_lat_conv.append(
                p_layer(
                    in_channels=self.out_channels // 4,
                    out_channels=self.out_channels // 4,
                    kernel_size=9,
                    padding=4,
                    weight_attr=ParamAttr(initializer=weight_attr),
                    bias_attr=False,
                )
            )

        self.incl1 = IntraCLBlock(
            self.out_channels // 4,
            reduce_factor=reduce_factor,
            intraclblock_config=intraclblock_config,
        )
        self.incl2 = IntraCLBlock(
            self.out_channels // 4,
            reduce_factor=reduce_factor,
            intraclblock_config=intraclblock_config,
        )
        self.incl3 = IntraCLBlock(
            self.out_channels // 4,
            reduce_factor=reduce_factor,
            intraclblock_config=intraclblock_config,
        )
        self.incl4 = IntraCLBlock(
            self.out_channels // 4,
            reduce_factor=reduce_factor,
            intraclblock_config=intraclblock_config,
        )

    def forward(self, x: List[paddle.Tensor]) -> paddle.Tensor:
        c2, c3, c4, c5 = x

        in5 = self.ins_conv[3](c5)
        in4 = self.ins_conv[2](c4)
        in3 = self.ins_conv[1](c3)
        in2 = self.ins_conv[0](c2)

        out4 = in4 + F.upsample(
            in5,
            scale_factor=2,
            mode=self.upsample_mode,
            align_mode=self.upsample_align_mode,
        )
        out3 = in3 + F.upsample(
            out4,
            scale_factor=2,
            mode=self.upsample_mode,
            align_mode=self.upsample_align_mode,
        )
        out2 = in2 + F.upsample(
            out3,
            scale_factor=2,
            mode=self.upsample_mode,
            align_mode=self.upsample_align_mode,
        )

        f5 = self.inp_conv[3](in5)
        f4 = self.inp_conv[2](out4)
        f3 = self.inp_conv[1](out3)
        f2 = self.inp_conv[0](out2)

        pan3 = f3 + self.pan_head_conv[0](f2)
        pan4 = f4 + self.pan_head_conv[1](pan3)
        pan5 = f5 + self.pan_head_conv[2](pan4)

        p2 = self.pan_lat_conv[0](f2)
        p3 = self.pan_lat_conv[1](pan3)
        p4 = self.pan_lat_conv[2](pan4)
        p5 = self.pan_lat_conv[3](pan5)

        p5 = self.incl4(p5)
        p4 = self.incl3(p4)
        p3 = self.incl2(p3)
        p2 = self.incl1(p2)

        p5 = F.upsample(
            p5,
            scale_factor=8,
            mode=self.upsample_mode,
            align_mode=self.upsample_align_mode,
        )
        p4 = F.upsample(
            p4,
            scale_factor=4,
            mode=self.upsample_mode,
            align_mode=self.upsample_align_mode,
        )
        p3 = F.upsample(
            p3,
            scale_factor=2,
            mode=self.upsample_mode,
            align_mode=self.upsample_align_mode,
        )

        fuse = paddle.concat([p5, p4, p3, p2], axis=1)
        return fuse


class ConvBNLayer(nn.Layer):
    """
    ConvBNLayer: Basic convolution + batch normalization + optional activation block

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int): Kernel size of convolution layer
        stride (int): Convolution stride
        padding (Union[int, str]): Padding value or mode (e.g. 'SAME') for convolution
        groups (int, optional): Number of grouped convolution groups, default is 1
        if_act (bool, optional): Whether to apply activation function, default is True
        act (Optional[str], optional): Activation function type ('relu' or 'hardswish'), default is None

    Returns:
        paddle.Tensor: Output tensor after convolution, BN and optional activation
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        padding: Union[int, str],
        groups: int = 1,
        if_act: bool = True,
        act: Optional[str] = None,
    ) -> None:
        super(ConvBNLayer, self).__init__()
        self.if_act = if_act
        self.act = act
        self.conv = nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=groups,
            bias_attr=False,
        )

        self.bn = nn.BatchNorm(num_channels=out_channels, act=None)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.conv(x)
        x = self.bn(x)
        if self.if_act:
            if self.act == "relu":
                x = F.relu(x)
            elif self.act == "hardswish":
                x = F.hardswish(x)
            else:
                print(
                    "The activation function({}) is selected incorrectly.".format(
                        self.act
                    )
                )
                exit()
        return x


class LocalModule(nn.Layer):
    """
    LocalModule: Feature enhancement module with concatenation and 1x1 projection

    Args:
        in_c (int): Number of input channels (before concatenation)
        mid_c (int): Number of intermediate channels for 3x3 ConvBNLayer
        act (str): Activation function type for ConvBNLayer

    Returns:
        paddle.Tensor: 1-channel output tensor after concatenation, conv and projection
    """

    def __init__(self, in_c: int, mid_c: int, act: str) -> None:
        super(self.__class__, self).__init__()
        self.last_3 = ConvBNLayer(in_c + 1, mid_c, 3, 1, 1, act=act)
        self.last_1 = nn.Conv2D(mid_c, 1, 1, 1, 0)

    def forward(self, x: paddle.Tensor, init_map: paddle.Tensor) -> paddle.Tensor:
        outf = paddle.concat([init_map, x], axis=1)
        out = self.last_1(self.last_3(outf))
        return out


class PFHeadLocal(DBHead):
    """
    PFHeadLocal: Enhanced DB head with local feature refinement for detection

    Args:
        in_channels (int): Number of input channels
        k (int): DB head hyperparameter (kernel factor)
        mode (str): Module size mode ('large' or 'small') to control intermediate channels
        scale_factor (int): Upsampling scale factor for feature maps
        act (str): Activation function type for LocalModule
        upsample_mode (str): Interpolation mode for upsample operation
        upsample_align_mode (int): Align mode for upsample operation
        **kwargs: Additional keyword arguments for parent DBHead class

    Returns:
        paddle.Tensor: Fused binarization map (average of base map and enhanced local map)
    """

    def __init__(
        self,
        in_channels: int,
        k: int,
        mode: str,
        scale_factor: int,
        act: str,
        upsample_mode: str,
        upsample_align_mode: int,
        **kwargs: Any,
    ) -> None:
        super(PFHeadLocal, self).__init__(in_channels, k, **kwargs)
        self.mode = mode

        self.up_conv = nn.Upsample(
            scale_factor=scale_factor,
            mode=upsample_mode,
            align_mode=upsample_align_mode,
        )

        if mode == "large":
            mid_ch = in_channels // 4
        elif mode == "small":
            mid_ch = in_channels // 8
        else:
            raise ValueError(f"mode must be 'large' or 'small', currently {mode}")
        self.cbn_layer = LocalModule(in_channels // 4, mid_ch, act)

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        base_maps, f = self.binarize(x, return_f=True)

        cbn_maps = self.cbn_layer(self.up_conv(f), base_maps)
        cbn_maps = F.sigmoid(cbn_maps)

        return 0.5 * (base_maps + cbn_maps)


class PPOCRV5ServerDet(BatchNormHFStateDictMixin, PretrainedModel):
    """
    PPOCRV5ServerDet: Server-side OCR detection model with PPHGNetV2, LKPAN and PFHeadLocal

    Args:
        config (PPOCRV5ServerDetConfig): Configuration object containing model hyperparameters

    Returns:
        List: List containing the detection output tensor (converted to numpy array on CPU)
    """

    config_class = PPOCRV5ServerDetConfig

    def __init__(self, config: PPOCRV5ServerDetConfig) -> None:
        super().__init__(config)

        self.upsample_mode = config.upsample_mode
        self.upsample_align_mode = config.upsample_align_mode
        self.backbone_stem_channels = config.backbone_stem_channels
        self.backbone_stage_config = config.backbone_stage_config
        self.backbone_use_lab = config.backbone_use_lab
        self.backbone_use_last_conv = config.backbone_use_last_conv
        self.backbone_class_expand = config.backbone_class_expand
        self.backbone_class_num = config.backbone_class_num
        self.backbone_lr_mult_list = config.backbone_lr_mult_list
        self.backbone_det = config.backbone_det
        self.backbone_out_indices = config.backbone_out_indices

        self.neck_out_channels = config.neck_out_channels
        self.neck_mode = config.neck_mode
        self.neck_reduce_factor = config.neck_reduce_factor
        self.neck_intraclblock_config = config.neck_intraclblock_config

        self.head_in_channels = config.head_in_channels
        self.head_k = config.head_k
        self.head_mode = config.head_mode
        self.head_scale_factor = config.head_scale_factor
        self.head_act = config.head_act
        self.head_kernel_list = config.head_kernel_list
        self.head_fix_nan = config.head_fix_nan

        self.backbone = PPHGNetV2(
            stage_config=self.backbone_stage_config,
            stem_channels=self.backbone_stem_channels,
            use_lab=self.backbone_use_lab,
            use_last_conv=self.backbone_use_last_conv,
            class_expand=self.backbone_class_expand,
            class_num=self.backbone_class_num,
            lr_mult_list=self.backbone_lr_mult_list,
            det=self.backbone_det,
            out_indices=self.backbone_out_indices,
        )

        neck_in_channels = self.backbone.out_channels
        self.neck = LKPAN(
            in_channels=neck_in_channels,
            out_channels=self.neck_out_channels,
            mode=self.neck_mode,
            reduce_factor=self.neck_reduce_factor,
            intraclblock_config=self.neck_intraclblock_config,
            upsample_mode=self.upsample_mode,
            upsample_align_mode=self.upsample_align_mode,
        )

        head_in_channels = self.neck.out_channels
        self.head = PFHeadLocal(
            in_channels=head_in_channels,
            k=self.head_k,
            mode=self.head_mode,
            scale_factor=self.head_scale_factor,
            act=self.head_act,
            upsample_mode=self.upsample_mode,
            upsample_align_mode=self.upsample_align_mode,
            kernel_list=self.head_kernel_list,
            fix_nan=self.head_fix_nan,
        )

    add_inference_operations("pp_ocrv5_server_det_forward")

    @benchmark.timeit_with_options(name="pp_ocrv5_server_det_forward")
    def forward(self, x: List) -> List:

        x = paddle.to_tensor(x[0])

        x = self.backbone(x)
        x = self.neck(x)
        x = self.head(x)

        return [x.cpu().numpy()]
